"""
Detectron2 is less polished than I thought it would be. It has two different
incompatible config systems. This file helps with the new LazyConfig system.
"""
from geowatch.tasks.detectron2._common import Detectron2WrapperBase
import ubelt as ub
import os


class Detectron2WrapperNewStyle(Detectron2WrapperBase):
    """
    Wrapper for the new-style detectron configs.

    CommandLine:
        xdoctest -m geowatch.tasks.detectron2._new_config_backend Detectron2WrapperNewStyle

    Example:
        >>> # xdoctest: +REQUIRES(module:fvcore)
        >>> from geowatch.tasks.detectron2.fit import DetectronFitCLI
        >>> from geowatch.tasks.detectron2._new_config_backend import *  # NOQA
        >>> import geowatch_tpl
        >>> geowatch_tpl.import_submodule('detectron2')  # NOQA
        >>> import kwcoco
        >>> import os
        >>> dset = kwcoco.CocoDataset.demo('vidshapes8')
        >>> dpath = ub.Path.appdir('geowatch/tests/detectron/fit_regnet')
        >>> config = DetectronFitCLI()
        >>> config['train_fpath'] = os.fspath(dset.fpath)
        >>> config['default_root_dir'] = os.fspath(dpath)
        >>> config['base'] = 'new_baselines/mask_rcnn_regnety_4gf_dds_FPN_400ep_LSJ.py'
        >>> config['init'] = 'new_baselines/mask_rcnn_regnety_4gf_dds_FPN_400ep_LSJ.py'
        >>> config['cfg'] = ub.codeblock(
                '''
                dataloader:
                    train:
                        total_batch_size: 4
                        num_workers: 1
                optimizer:
                    lr: 0.001
                train:
                    amp:
                        enabled: true
                    max_iter: 10
                    eval_period: 1
                    log_period: 1
                    checkpointer:
                        period: 1
                        max_to_keep: 100
                    device: cuda
                ''')
        >>> self = Detectron2WrapperNewStyle(config)
        >>> self.register_datasets()
        >>> self.resolve_config()
        >>> self.build_trainer()
        >>> self.train()
    """

    def __init__(self, config):
        super().__init__(config)

    def resolve_config(self):
        """
        I really dont like these "new" lazy Python configs.
        There isnt a clear way to create an instance of the model.

        References:
            https://github.com/facebookresearch/detectron2/blob/main/docs/tutorials/lazyconfigs.md

        Ignore:
            from detectron2 import model_zoo
            cfg = model_zoo.get_config('new_baselines/mask_rcnn_regnety_4gf_dds_FPN_400ep_LSJ.py')
            instantiate(cfg.model)
        """

        import omegaconf
        import rich
        from rich.markup import escape
        from detectron2 import model_zoo
        # from omegaconf import OmegaConf
        import kwutil
        config = self.config
        dataset_infos = self.dataset_infos

        @ub.urepr.extensions.register(omegaconf.dictconfig.DictConfig)
        def _format_omegaconfg(d, **kw):
            # Hack for printing omegaconfig with some reasonablness
            return ub.util_repr._format_dict(d, **kw)[0]

        # # Hack for hashing
        # found = None
        # for k, v in ub.hash_data.extensions._hash_dispatch.registry.items():
        #     if k.__name__ == 'dict':
        #         found = v
        # assert found is not None
        # ub.hash_data.extensions._hash_dispatch.register(omegaconf.dictconfig.DictConfig)(found)

        cfg = model_zoo.get_config(config.base)
        cfg.dataloader.train.dataset.names = dataset_infos['train']['name']

        if dataset_infos['vali'] is None:
            cfg.dataloader.test.dataset.names = None
        else:
            cfg.dataloader.test.dataset.names = dataset_infos['vali']['name']

        base_cfg_text = omegaconf_dumps(cfg)
        rich.print(f'base cfg = {escape(base_cfg_text)}')

        cfg.model.roi_heads.num_classes = len(dataset_infos['train']['categories'])

        # Override config values
        cfg_final_layer = kwutil.Yaml.coerce(config.cfg, backend='pyyaml')
        walker = ub.IndexableWalker(cfg)
        to_set_walker = ub.IndexableWalker(cfg_final_layer)
        for p, v in to_set_walker:
            if not isinstance(v, dict):
                walker[p] = v

        if self.config.init == 'noop':
            cfg.train.init_checkpoint = ''
        else:
            cfg.train.init_checkpoint = model_zoo.get_checkpoint_url(self.config.init)  # Let training initialize from model zoo

        cfg.train.output_dir = None  # hack: null out for the initial hashing
        text = omegaconf_dumps(cfg)
        hashid = ub.hash_data(text)[0:8]

        output_dpath = ub.Path(config.default_root_dir) / f'v_{hashid}'
        output_dpath.ensuredir()
        self.output_dpath = output_dpath
        cfg.train.output_dir = os.fspath(output_dpath)
        overlaid_cfg_text = omegaconf_dumps(cfg)
        rich.print(f'overlaid cfg =  {escape(overlaid_cfg_text)}')
        self.cfg = cfg

    def dump_model_config(self):
        from detectron2.config import LazyConfig
        output_dpath = self.output_dpath.ensuredir()
        LazyConfig.save(self.cfg, os.fspath(output_dpath / 'detectron_config.yaml'))

    def build_predictor(self):
        ...

    def build_trainer(self):
        from detectron2.checkpoint import DetectionCheckpointer
        from detectron2.config import instantiate
        from detectron2.engine import (
            AMPTrainer,
            SimpleTrainer,
            default_writers,
            hooks,
        )
        from detectron2.engine.defaults import create_ddp_model
        from detectron2.utils import comm
        """
        cfg: an object with the following attributes:
            model: instantiate to a module
            dataloader.{train,test}: instantiate to dataloaders
            dataloader.evaluator: instantiate to evaluator for test set
            optimizer: instantaite to an optimizer
            lr_multiplier: instantiate to a fvcore scheduler
            train: other misc config defined in `configs/common/train.py`, including:
                output_dir (str)
                init_checkpoint (str)
                amp.enabled (bool)
                max_iter (int)
                eval_period, log_period (int)
                device (str)
                checkpointer (dict)
                ddp (dict)
        """
        cfg = self.cfg.copy()
        model = instantiate(cfg.model)
        model.to(cfg.train.device)

        cfg.optimizer.params.model = model
        optim = instantiate(cfg.optimizer)

        train_loader = instantiate(cfg.dataloader.train)

        model = create_ddp_model(model, **cfg.train.ddp)
        trainer = (AMPTrainer if cfg.train.amp.enabled else SimpleTrainer)(model, train_loader, optim)
        checkpointer = DetectionCheckpointer(
            model,
            cfg.train.output_dir,
            trainer=trainer,
        )
        trainer.register_hooks([
            hooks.IterationTimer(),
            hooks.LRScheduler(scheduler=instantiate(cfg.lr_multiplier)),
            (
                hooks.PeriodicCheckpointer(checkpointer, **cfg.train.checkpointer)
                if comm.is_main_process()
                else None
            ),
        ])
        trainer.register_hooks([
            (None if self.dataset_infos['vali'] is None else hooks.EvalHook(cfg.train.eval_period, lambda: do_test(cfg, model))),
        ])
        trainer.register_hooks([
            (
                hooks.PeriodicWriter(
                    default_writers(cfg.train.output_dir, cfg.train.max_iter),
                    period=cfg.train.log_period,
                )
                if comm.is_main_process()
                else None
            ),
        ]
        )

        checkpointer.resume_or_load(cfg.train.init_checkpoint)
        if checkpointer.has_checkpoint():
            # The checkpoint stores the training iteration that just finished, thus we start
            # at the next iteration
            self.start_iter = trainer.iter + 1
        else:
            self.start_iter = 0

        self.trainer = trainer

    def train(self):
        self.dump_model_config()
        if self.trainer is None:
            self.build_trainer()
        import rich
        rich.print(f'Output Dpath: [link={self.output_dpath}]{self.output_dpath}[/link]')
        self.trainer.train(self.start_iter, self.cfg.train.max_iter)
        rich.print(f'Output Dpath: [link={self.output_dpath}]{self.output_dpath}[/link]')


def do_test(cfg, model):
    from detectron2.config import instantiate
    from detectron2.evaluation import inference_on_dataset, print_csv_format
    if "evaluator" in cfg.dataloader:
        ret = inference_on_dataset(
            model,
            instantiate(cfg.dataloader.test),
            instantiate(cfg.dataloader.evaluator),
        )
        print_csv_format(ret)
        return ret


def omegaconf_dumps(cfg):
    """
    Get some semi-serializable representation of the new-style configuration
    """
    from omegaconf import OmegaConf, SCMode
    dict = OmegaConf.to_container(
        cfg,
        # Do not resolve interpolation when saving, i.e. do not turn ${a} into
        # actual values when saving.
        resolve=False,
        # Save structures (dataclasses) in a format that can be instantiated later.
        # Without this option, the type information of the dataclass will be erased.
        structured_config_mode=SCMode.INSTANTIATE,
    )
    import yaml
    dumped = yaml.dump(dict, default_flow_style=None, allow_unicode=True, width=9999)
    import io
    f = io.StringIO()
    f.write(dumped)
    text = f.getvalue()
    return text



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Squeeze and excite layer.

This script implements the squeeze-and-excite (SE), proposed in
- Squeeze-and-Excitation Networks, Jie Hu, Li Shen, Samuel Albanie,
Gang Sun, Enhua Wu. In CVPR 2018.

Recently, this SE operation is further simplied with a single fully
connected layer, referred as simplified_squeeze_and_excite in our
implementation. For details, please see
- Lee and Park proposed to use only one fully connected layer in SE.
CenterMask : Real-Time Anchor-Free Instance Segmentation.
Youngwan Lee and Jongyoul Park. In CVPR 2020.
"""
from typing import Optional

import tensorflow as tf
from absl import logging
from deeplab2.model import utils
from deeplab2.model.layers import activations

layers = tf.keras.layers


class SimplifiedSqueezeAndExcite(tf.keras.layers.Layer):
    """A simplified squeeze-and-excite layer.

    Original squeeze-and-exciation (SE) is proposed in
    Squeeze-and-Excitation Networks, Jie Hu, Li Shen, Samuel Albanie,
    Gang Sun, Enhua Wu. In CVPR 2018.

    Lee and Park proposed to use only one fully connected layer in SE.
    CenterMask : Real-Time Anchor-Free Instance Segmentation.
    Youngwan Lee and Jongyoul Park. In CVPR 2020.

    In this function, we implement the simplified version of SE.

    Additionally, we follow MobileNetv3 to use the hard sigmoid function.
    """

    def __init__(self, squeeze_channels, name=None):
        """Initializes a simplified squeeze-and-excite layer.

        Args:
          squeeze_channels: Integer, channels for the squeezed features.
          name: An optional string specifying the operation name.
        """
        super(SimplifiedSqueezeAndExcite, self).__init__(name=name)
        self._squeeze_channels = squeeze_channels

        self._se_conv = layers.Conv2D(self._squeeze_channels,
                                      1,
                                      name='squeeze_and_excite',
                                      use_bias=True,
                                      kernel_initializer='VarianceScaling')
        self._hard_sigmoid = activations.get_activation('hard_sigmoid')

    def call(self, input_tensor):
        """Performs a forward pass.

        Args:
          input_tensor: An input tensor of type tf.Tensor with shape [batch, height,
            width, channels].

        Returns:
          The output tensor.
        """
        pooled = tf.reduce_mean(input_tensor, [1, 2], keepdims=True)
        squeezed = self._se_conv(pooled)
        excited = self._hard_sigmoid(squeezed) * input_tensor
        return excited

    def get_config(self):
        config = {
            'squeeze_channels': self._squeeze_channels,
        }
        base_config = super(SimplifiedSqueezeAndExcite, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class SqueezeAndExcite(tf.keras.layers.Layer):
    """Creates a squeeze and excitation layer.

    Reference: Squeeze-and-Excitation Networks, Jie Hu, Li Shen, Samuel Albanie,
    Gang Sun, Enhua Wu. In CVPR 2018.
    This implementation follows the original SE and differs from the above
    simplified version.
    """

    def __init__(
            self,
            in_filters: int,
            out_filters: int,
            se_ratio: float,
            divisible_by: int = 1,
            kernel_initializer: str = 'VarianceScaling',
            kernel_regularizer: Optional[tf.keras.regularizers.Regularizer] = None,
            bias_regularizer: Optional[tf.keras.regularizers.Regularizer] = None,
            activation: str = 'relu',
            gating_activation: str = 'sigmoid',
            name: Optional[str] = None):
        """Initializes a squeeze and excitation layer.

        Args:
          in_filters: The number of filters that se_ratio should be applied to.
          out_filters: The number of filters of the output tensor.
          se_ratio: The SE ratio for the squeeze and excitation layer.
          divisible_by: An `int` that ensures all inner dimensions are divisible by
            this number.
          kernel_initializer: The kernel_initializer for convolutional
            layers.
          kernel_regularizer: A `tf.keras.regularizers.Regularizer` object for
            Conv2D. Default to None.
          bias_regularizer: A `tf.keras.regularizers.Regularizer` object for Conv2d.
            Default to None.
          activation: The name of the activation function.
          gating_activation: The name of the activation function for final
            gating function.
          name: The layer name.
        """
        super(SqueezeAndExcite, self).__init__(name=name)

        self._in_filters = in_filters
        self._out_filters = out_filters
        self._se_ratio = se_ratio
        self._divisible_by = divisible_by
        self._activation = activation
        self._gating_activation = gating_activation
        self._kernel_initializer = kernel_initializer
        self._kernel_regularizer = kernel_regularizer
        self._bias_regularizer = bias_regularizer
        if tf.keras.backend.image_data_format() == 'channels_last':
            self._spatial_axis = [1, 2]
        else:
            self._spatial_axis = [2, 3]
        self._activation_fn = activations.get_activation(activation)
        self._gating_activation_fn = activations.get_activation(gating_activation)

        num_reduced_filters = utils.make_divisible(
            max(1, int(self._in_filters * self._se_ratio)),
            divisor=self._divisible_by)
        if self._se_ratio > 1.0:
            logging.warn('Squeezing ratio %d is larger than 1.0.', self._se_ratio)

        self._se_reduce = tf.keras.layers.Conv2D(
            filters=num_reduced_filters,
            kernel_size=1,
            strides=1,
            padding='same',
            use_bias=True,
            kernel_initializer=self._kernel_initializer,
            kernel_regularizer=self._kernel_regularizer,
            bias_regularizer=self._bias_regularizer,
            name=name + '_reduce')

        self._se_expand = tf.keras.layers.Conv2D(
            filters=self._out_filters,
            kernel_size=1,
            strides=1,
            padding='same',
            use_bias=True,
            kernel_initializer=self._kernel_initializer,
            kernel_regularizer=self._kernel_regularizer,
            bias_regularizer=self._bias_regularizer,
            name=name + '_expand')

    def call(self, inputs):
        x = tf.reduce_mean(inputs, self._spatial_axis, keepdims=True)
        x = self._activation_fn(self._se_reduce(x))
        x = self._gating_activation_fn(self._se_expand(x))
        return x * inputs



#!/usr/bin/env python3
"""
PCD = parallel change detection
"""

import os
import ubelt as ub
import scriptconfig as scfg


class ScoreTracksConfig(scfg.DataConfig):
    """
    Filter tracks based on the depth detector.
    """
    input_kwcoco = scfg.Value(None, required=True, help=ub.paragraph(
        '''
        The input kwcoco file with the high-resolution images to use for the
        depth filter. This does not to cover all sites, any site this does not
        cover will be automatically accepted.
        '''), position=1, alias=['in_file'], group='inputs')

    model_fpath = scfg.Value(None, help=ub.paragraph(
        '''
        Path to the depth_pcd site validation model
        '''), group='inputs')

    input_region = scfg.Value(None, help=ub.paragraph(
        '''
        The coercable input region model with site summaries
        '''), group='inputs')

    out_kwcoco = scfg.Value(None, help=ub.paragraph(
        '''
            The file path to write the "tracked" kwcoco file to.
            '''), group='outputs')


def score_tracks(img_coco_dset, model_fpath):
    from geowatch.tasks.depth_pcd.model import getModel, normalize, TPL_DPATH

    import numpy as np
    import cv2
    import kwimage
    import ndsampler
    import pandas as pd
    from tqdm import tqdm
    import warnings

    print('loading site validation model')
    proto_fpath = TPL_DPATH / 'deeplab2/max_deeplab_s_backbone_os16.textproto'
    use_ln = False
    res = '2GSD'
    if str(model_fpath).find('model4') != -1:
        use_ln = True
        res = '1GSD'
    model = getModel(proto=proto_fpath, use_ln=use_ln)

    model.load_weights(model_fpath, by_name=True, skip_mismatch=True)
    # model.load_weights('/media/hdd2/antonio/models/urbanTCDs-use.h5', by_name=True, skip_mismatch=True)

    to_keep = []
    for coco_img in img_coco_dset.images().coco_images:
        if coco_img['sensor_coarse'] == 'WV' and 'red' in coco_img.channels:
            to_keep.append(coco_img['id'])

    dset = img_coco_dset.subset(to_keep)

    sampler = ndsampler.CocoSampler(dset)

    all_videos = dset.videos()
    all_annots = dset.annots()

    if len(all_annots) == 0:
        print("Nothing to filter")
        return img_coco_dset

    vidid_to_name = all_videos.lookup('name', keepid=True)
    # vidname_to_id = ub.udict(vidid_to_name).invert()
    annot_video_ids = all_annots.images.lookup('video_id')
    annot_timestamp = all_annots.images.lookup('timestamp')
    annot_video_names = list(ub.take(vidid_to_name, annot_video_ids))
    annot_image_ids = all_annots.lookup('image_id')
    annot_track_ids = all_annots.lookup('track_id')

    annot_df = pd.DataFrame({
        'video_id': np.array(annot_video_ids),
        'timestamp': np.array(annot_timestamp),
        'video_name': np.array(annot_video_names),
        'image_id': np.array(annot_image_ids),
        'track_id': np.array(annot_track_ids),
        'id': np.array(all_annots.ids),
    })
    # Group by track and video name.
    trackid_to_group = dict(list(annot_df.groupby('track_id')))

    # TODO: use new kwcoco track mechanisms
    # Ideally the dataset will be passed to us with tracks.
    # This will happen once reproject annotations handles it.
    if 'tracks' not in img_coco_dset.dataset.keys():
        img_coco_dset.dataset['tracks'] = []
    tracks = img_coco_dset.dataset['tracks']
    tq = tqdm(total=len(trackid_to_group))

    for track_id, orig_track_group in trackid_to_group.items():
        # TODO: use new kwcoco track mechanisms
        track_obj = {
            'id': track_id,
            'name': track_id,  # add a name for "future-proofing"
            'score': 1.0,
            'src': 'sv_depth_pcd'
        }
        tracks.append(track_obj)
        # Does the track appear in more than one video?
        video_names = orig_track_group['video_name'].unique()
        if len(video_names) > 1:
            if track_id not in video_names:
                msg = ub.paragraph(
                    f'''
                    track-id {track_id} expected to correspond with video names
                    'in site-cropped datasets
                    ''')
                warnings.warn(msg)
                continue
            # take the "main" video for this track
            track_group = orig_track_group[orig_track_group['video_name'] == track_id]
        else:
            track_group = orig_track_group

        track_group = track_group.sort_values('timestamp')

        video_id = track_group["video_id"].iloc[0]
        video_name = track_group["video_name"].iloc[0]
        image_ids = track_group["image_id"].tolist()
        first_annot_id = track_group["id"].iloc[0]
        first_image_id = image_ids[0]
        first_coco_img = img_coco_dset.coco_image(first_image_id)
        first_annot = img_coco_dset.anns[first_annot_id]

        # Read the location of the first annotation in "image space"
        imgspace_annot_box = kwimage.Box.coerce(first_annot['bbox'], format='xywh')

        # Convert the location of the annotation to "video space"
        vidspace_annot_box = imgspace_annot_box.warp(first_coco_img.warp_vid_from_img)

        # Because we want a higher resolution, we need to scale the requested
        # videospace region down. Looks like quantization errors may happen
        # here not sure how I deal with in the dataloader, it probably needs to
        # be fixed there too.
        scale_res_from_vidspace = first_coco_img._scalefactor_for_resolution(space='video', resolution=res)

        # This is the transform from video space (i.e. the space we use to talk
        # to ndsampler) to the final resolution we want to sample.
        warp_res_from_vidspace = kwimage.Affine.scale(scale_res_from_vidspace)

        # Convert the video space annotation into requested resolution "window space"
        winspace_annot_box = vidspace_annot_box.warp(warp_res_from_vidspace)

        # Convert to center xy/with/height format
        winspace_annot_box = winspace_annot_box.toformat('cxywh')

        # Force the box to be a specific size at our window resolution
        force_dsize = (224, 224)
        winspace_target_box = winspace_annot_box.resize(*force_dsize)

        # Convert the box back to videospace
        vidspace_target_box = winspace_target_box.warp(warp_res_from_vidspace.inv())

        # Get the slice for video space
        # vidspace_slice = vidspace_target_box.quantize().to_slice()
        # vidspace_slice = vidspace_target_box.quantize().to_ltrb().quantize().to_slice()
        vidspace_slice = vidspace_target_box.to_ltrb().quantize().to_slice()

        # The space slice is specified in video space, so to recover the
        # requested resolution, we pass the videospace -> samplespace
        # scalefactor. We could assert that decompose opertion should have
        # zeros everywhere but scale, but we aren't.
        scale = warp_res_from_vidspace.decompose()['scale']

        target = {
            'vidid': video_id,
            'gids': image_ids,
            'channels': 'blue|green|red',
            'allow_augment': False,
            'space_slice': vidspace_slice,

            'use_native_scale': False,
            'scale': scale,
        }
        try:
            data = sampler.load_sample(target, with_annots=False)
        except ValueError:
            print('warning: failed to sample with value error')
            tq.update(1)
            continue
        except Exception:
            # Sample again with more info or debugging and then
            target['verbose_ndsample'] = True
            sampler.load_sample(target, with_annots=False)
            print('warning: failed to sample with unknown exception')
            tq.update(1)
            continue
            # raise

        ims = data['im']

        good_ims = []
        for i in ims:
            im = np.stack([ii[..., 0] for ii in i], axis=-1) / (4096 * 2)
            im = im.clip(0, 1)
            im = (im * 255).astype(np.uint8)
            if im.shape[-1] == 3:
                im = cv2.cvtColor(im, cv2.COLOR_BGR2GRAY)
            if np.mean(im == 0) > .2:
                continue
            norm_im = normalize(im)
            good_ims.append(norm_im)

        # a little average at start vs end
        nAvg = 2
        if len(good_ims) < nAvg + 1:
            tq.update(1)
            continue
        ims = []
        for i in range(nAvg):
            for j in range(-1, -nAvg - 1, -1):
                first = good_ims[i]
                last = good_ims[j]
                ims.append(np.stack([first, 0.5 * first + 0.5 * last, last], axis=-1).astype(np.float32))

        score = np.mean(model.predict(np.array(ims), batch_size=1, verbose=False)[8])
        track_obj['score'] = float(score)
        tq.set_description(f'{video_name} score {score:3.2f}')
        tq.update(1)
        if 0:

            cv2.namedWindow('main', cv2.WINDOW_GUI_NORMAL)
            i = 1
            first = good_ims[0]
            last = good_ims[-1]
            while 1:
                if i == 1:
                    i = 0
                    im = first
                else:
                    i = 1
                    im = last
                cv2.imshow('main', im / 255)
                q = cv2.waitKey(0)
                if q == 113:
                    break

        #        ks = list(coco_dset.index.videos.keys())

    return img_coco_dset


def main(cmdline=True, **kwargs):
    args = ScoreTracksConfig.cli(cmdline=cmdline, data=kwargs, strict=True)
    import rich
    rich.print('args = {}'.format(ub.urepr(args, nl=1)))

    # Import this first
    print('Importing tensorflow stuff (can take a sec)')
    from geowatch.tasks.depth_pcd.model import getModel, normalize, TPL_DPATH  # NOQA

    import kwcoco
    from kwcoco.util import util_json
    from geowatch.utils import process_context

    if args.model_fpath is None:
        print('warning: the path to the model was not explicitly specified, '
              'attempting to automatically infer it')
        cand_model_path = ub.Path(os.environ.get('DVC_EXPT_DPATH', '')) / 'models/depth_pcd/basicModel2.h5'
        if cand_model_path.exists():
            args.model_fpath = cand_model_path
        else:
            raise IOError(
                f'Attempted to infer model path {cand_model_path}, '
                'but it does not exist. Please specify it explicitly')

    model_fpath = ub.Path(args.model_fpath)
    if not model_fpath.exists():
        raise IOError(f'Specified {model_fpath=} does not exist')

    # Args will be serailized in kwcoco, so make sure it can be coerced to json
    jsonified_config = util_json.ensure_json_serializable(args.asdict())
    walker = ub.IndexableWalker(jsonified_config)
    for problem in util_json.find_json_unserializable(jsonified_config):
        bad_data = problem['data']
        walker[problem['loc']] = str(bad_data)

    proc_context = process_context.ProcessContext(
        name='geowatch.tasks.depth_pcd.score_tracks', type='process',
        config=jsonified_config,
        track_emissions=False,
    )
    proc_context.start()

    img_coco_dset = kwcoco.CocoDataset.coerce(args.input_kwcoco)

    # Project the site polygons onto the kwcoco dataset.
    from geowatch.cli import reproject_annotations
    img_coco_dset = reproject_annotations.main(
        cmdline=0, src=img_coco_dset,
        dst='return',
        region_models=args.input_region,
        status_to_catname={'system_confirmed': 'positive'},
        role='pred_poly',
        validate_checks=False,
        clear_existing=False,
    )

    coco_dset = score_tracks(img_coco_dset, model_fpath)

    proc_context.stop()
    out_kwcoco = args.out_kwcoco
    if out_kwcoco is not None:
        coco_dset = coco_dset.reroot(absolute=True, check=False)
        # Add tracking audit data to the kwcoco file
        coco_info = coco_dset.dataset.get('info', [])
        coco_info.append(proc_context.obj)
        coco_dset.fpath = out_kwcoco
        ub.Path(out_kwcoco).parent.ensuredir()
        print(f'write to coco_dset.fpath={coco_dset.fpath}')
        coco_dset.dump(out_kwcoco, indent=2)


r'''
Example:

    ### Run BAS and then run SV on top of it.

    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
    DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware=auto)
    BAS_MODEL_FPATH=$DVC_EXPT_DPATH/models/fusion/Drop6-MeanYear10GSD-V2/packages/Drop6_TCombo1Year_BAS_10GSD_V2_landcover_split6_V47/Drop6_TCombo1Year_BAS_10GSD_V2_landcover_split6_V47_epoch47_step3026.pt

    # Predict BAS Heatmaps
    python -m geowatch.tasks.fusion.predict \
        --package_fpath="$BAS_MODEL_FPATH" \
        --test_dataset=$DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/combo_imganns-KR_R002_I2L.kwcoco.zip \
        --pred_dataset=$DVC_EXPT_DPATH/_test_dzyne_sv/pred_heatmaps.kwcoco.zip \
        --chip_overlap="0.3" \
        --chip_dims="196,196" \
        --time_span="auto" \
        --fixed_resolution="10GSD" \
        --time_sampling="soft4" \
        --drop_unused_frames="True"  \
        --num_workers="4" \
        --devices="0," \
        --batch_size="1" \
        --with_saliency="True" \
        --with_class="False" \
        --with_change="False"

    # Convert Heatmaps to Polygons
    python -m geowatch.cli.run_tracker \
        --in_file "$DVC_EXPT_DPATH/_test_dzyne_sv/pred_heatmaps.kwcoco.zip" \
        --default_track_fn saliency_heatmaps \
        --track_kwargs '{
            "agg_fn": "probs",
            "thresh": 0.4,
            "time_thresh": 0.8,
            "inner_window_size": "1y",
            "inner_agg_fn": "mean",
            "norm_ord": "inf",
            "resolution": "10GSD",
            "moving_window_size": null,
            "poly_merge_method": "v2",
            "polygon_simplify_tolerance": 1,
            "min_area_square_meters": 7200,
            "max_area_square_meters": 8000000
        }' \
        --clear_annots=True \
        --site_summary 'None' \
        --boundary_region $DVC_DATA_DPATH/annotations/drop6/region_models \
        --out_site_summaries_fpath "$DVC_EXPT_DPATH/_test_dzyne_sv/site_summaries_manifest.json" \
        --out_site_summaries_dir "$DVC_EXPT_DPATH/_test_dzyne_sv/site_summaries" \
        --out_sites_fpath "$DVC_EXPT_DPATH/_test_dzyne_sv/sites_manifest.json" \
        --out_sites_dir "$DVC_EXPT_DPATH/_test_dzyne_sv/sites" \
        --out_kwcoco "$DVC_EXPT_DPATH/_test_dzyne_sv/poly.kwcoco.zip"

    # Score the Initial Predictions
    python -m geowatch.cli.run_metrics_framework \
        --merge=True \
        --name "todo" \
        --true_site_dpath "$DVC_DATA_DPATH/annotations/drop6/site_models" \
        --true_region_dpath "$DVC_DATA_DPATH/annotations/drop6/region_models" \
        --pred_sites "$DVC_EXPT_DPATH/_test_dzyne_sv/sites_manifest.json" \
        --tmp_dir "$DVC_EXPT_DPATH/_test_dzyne_sv/eval_before/tmp" \
        --out_dir "$DVC_EXPT_DPATH/_test_dzyne_sv/eval_before" \
        --merge_fpath "$DVC_EXPT_DPATH/_test_dzyne_sv/eval_before/poly_eval_before.json"

    # Run the Site Validation Filter
    python -m geowatch.tasks.depth_pcd.score_tracks \
        --input_kwcoco $DVC_DATA_DPATH/Drop6/imgonly-KR_R002.kwcoco.json \
        --input_region "$DVC_EXPT_DPATH/_test_dzyne_sv/site_summaries_manifest.json" \
        --input_sites "$DVC_EXPT_DPATH/_test_dzyne_sv/sites_manifest.json" \
        --model_fpath $DVC_EXPT_DPATH/models/depth_pcd/basicModel2.h5 \
        --out_kwcoco "$DVC_EXPT_DPATH/_test_dzyne_sv/filtered_poly.kwcoco.zip" \


        --output_sites_dpath  "$DVC_EXPT_DPATH/_test_dzyne_sv/filtered_sites" \
        --output_region_fpath "$DVC_EXPT_DPATH/_test_dzyne_sv/filtered_site_summaries.json" \
        --output_site_manifest_fpath  "$DVC_EXPT_DPATH/_test_dzyne_sv/filtered_sites_manifest.json" \
        --threshold 0.4

    # Score the Filtered Predictions
    python -m geowatch.cli.run_metrics_framework \
        --merge=True \
        --name "todo" \
        --true_site_dpath "$DVC_DATA_DPATH/annotations/drop6/site_models" \
        --true_region_dpath "$DVC_DATA_DPATH/annotations/drop6/region_models" \
        --pred_sites "$DVC_EXPT_DPATH/_test_dzyne_sv/filtered_sites_manifest.json" \
        --tmp_dir "$DVC_EXPT_DPATH/_test_dzyne_sv/eval_after/tmp" \
        --out_dir "$DVC_EXPT_DPATH/_test_dzyne_sv/eval_after" \
        --merge_fpath "$DVC_EXPT_DPATH/_test_dzyne_sv/eval_after/poly_eval_after.json"

    python -c "if 1:
        import pandas as pd
        import rich
        import json
        import ubelt as ub
        text1 = ub.Path('$DVC_EXPT_DPATH/_test_dzyne_sv/eval_before/poly_eval_before.json').read_text()
        text2 = ub.Path('$DVC_EXPT_DPATH/_test_dzyne_sv/eval_after/poly_eval_after.json').read_text()
        data1 = json.loads(text1)
        data2 = json.loads(text2)
        df1 = pd.read_json(json.dumps(data1['best_bas_rows']), orient='table')
        df2 = pd.read_json(json.dumps(data2['best_bas_rows']), orient='table')
        print('BEFORE:')
        rich.print(df1)
        print('After:')
        rich.print(df2)
    "


Example in MLOPs:

    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
    DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware=auto)
    geowatch schedule --params="
        matrix:
            bas_pxl.package_fpath:
                - $DVC_EXPT_DPATH/models/fusion/Drop6-MeanYear10GSD-V2/packages/Drop6_TCombo1Year_BAS_10GSD_V2_landcover_split6_V47/Drop6_TCombo1Year_BAS_10GSD_V2_landcover_split6_V47_epoch47_step3026.pt
            bas_pxl.test_dataset:
                - $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/combo_imganns-KR_R001_I2LS.kwcoco.zip
                # - $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/combo_imganns-KR_R002_I2LS.kwcoco.zip
                # - $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/combo_imganns-BR_R002_I2LS.kwcoco.zip
                # - $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/combo_imganns-CH_R001_I2LS.kwcoco.zip
                # - $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/combo_imganns-NZ_R001_I2LS.kwcoco.zip
                # - $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/combo_imganns-AE_R001_I2LS.kwcoco.zip
            bas_pxl.chip_overlap: 0.3
            bas_pxl.chip_dims: auto
            bas_pxl.time_span: auto
            bas_pxl.time_sampling: soft4
            bas_poly.thresh:
                # - 0.42
                - 0.425
            bas_poly.inner_window_size: 1y
            bas_poly.inner_agg_fn: mean
            bas_poly.norm_ord: inf
            bas_poly.polygon_simplify_tolerance: 1
            bas_poly.agg_fn: probs
            bas_poly.time_thresh:
                # - 0.65
                - 0.8
            bas_poly.resolution: 10GSD
            bas_poly.moving_window_size: null
            bas_poly.poly_merge_method: 'v2'
            bas_poly.min_area_square_meters: 7200
            bas_poly.max_area_square_meters: 8000000
            bas_poly.boundary_region: $DVC_DATA_DPATH/annotations/drop6/region_models
            bas_poly_eval.true_site_dpath: $DVC_DATA_DPATH/annotations/drop6/site_models
            bas_poly_eval.true_region_dpath: $DVC_DATA_DPATH/annotations/drop6/region_models
            bas_pxl.enabled: 1
            bas_pxl_eval.enabled: 0
            bas_poly.enabled: 1
            bas_poly_eval.enabled: 1
            bas_poly_viz.enabled: 0
            sv_crop.enabled: 1
            sv_crop.minimum_size: "256x256@2GSD"
            sv_crop.num_start_frames: 10
            sv_crop.num_end_frames: 10
            sv_crop.context_factor: 1.5

            sv_dino_boxes.enabled: 1
            sv_dino_boxes.package_fpath: $DVC_EXPT_DPATH/models/kitware/xview_dino.pt
            sv_dino_boxes.window_dims: 256
            sv_dino_boxes.window_overlap: 0.5
            sv_dino_boxes.fixed_resolution: 3GSD

            sv_dino_filter.enabled: 1
            sv_dino_filter.end_min_score: 0.15
            sv_dino_filter.start_max_score: 1.0
            sv_dino_filter.box_score_threshold: 0.01
            sv_dino_filter.box_isect_threshold: 0.1

            sv_depth_score.enabled: 1
            sv_depth_score.model_fpath: $DVC_EXPT_DPATH/models/depth_pcd/basicModel2.h5
            sv_depth_filter.threshold:
                - 0.20
            #     - 0.4
        submatrices:
            - bas_pxl.test_dataset: $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/combo_imganns-KR_R001_I2LS.kwcoco.zip
              sv_crop.crop_src_fpath: $DVC_DATA_DPATH/Drop6/imgonly-KR_R001.kwcoco.json
            - bas_pxl.test_dataset: $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/combo_imganns-KR_R002_I2LS.kwcoco.zip
              sv_crop.crop_src_fpath: $DVC_DATA_DPATH/Drop6/imgonly-KR_R002.kwcoco.json
            - bas_pxl.test_dataset: $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/combo_imganns-BR_R002_I2LS.kwcoco.zip
              sv_crop.crop_src_fpath: $DVC_DATA_DPATH/Drop6/imgonly-BR_R002.kwcoco.json
            - bas_pxl.test_dataset: $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/combo_imganns-CH_R001_I2LS.kwcoco.zip
              sv_crop.crop_src_fpath: $DVC_DATA_DPATH/Drop6/imgonly-CH_R001.kwcoco.json
            - bas_pxl.test_dataset: $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/combo_imganns-NZ_R001_I2LS.kwcoco.zip
              sv_crop.crop_src_fpath: $DVC_DATA_DPATH/Drop6/imgonly-NZ_R001.kwcoco.json
            - bas_pxl.test_dataset: $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/combo_imganns-AE_R001_I2LS.kwcoco.zip
              sv_crop.crop_src_fpath: $DVC_DATA_DPATH/Drop6/imgonly-AE_R001.kwcoco.json
        " \
        --root_dpath="$DVC_EXPT_DPATH/_mlops_test_depth_pcd2" \
        --devices="0," --tmux_workers=2 \
        --backend=serial --queue_name "_mlops_test_depth_pcd2" \
        --pipeline=bas_building_and_depth_vali \
        --skip_existing=1 \
        --run=0

        --pipeline=bas_depth_vali \


DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware=auto)
geowatch aggregate \
    --pipeline=bas_building_vali \
    --target \
        "$DVC_EXPT_DPATH/_mlops_test_depth_pcd" \
    --stdout_report="
        top_k: 100
        per_group: 2
        macro_analysis: 0
        analyze: 0
        reference_region: final
        print_models: True
    " \
    --resource_report=0 \
    --query='
        `params.bas_poly.thresh` == 0.425 and
        `params.bas_pxl.package_fpath`.str.contains("V47_epoch47_")
    ' \
    --plot_params="
        enabled: False
        compare_sv_hack: True
        stats_ranking: 0
        min_variations: 1
    " \
    --export_tables=0 \
    --io_workers=10 \
    --output_dpath="$DVC_EXPT_DPATH/_mlops_test_depth_pcd/aggregate" \
    --rois=KR_R002,CH_R001,NZ_R001

    # --rois=KR_R002,

    CH_R001,NZ_R001,KR_R001,BR_R002
'''
if __name__ == '__main__':
    main()



"""
Original model code "liberated" from https://github.com/toandaominh1997/EfficientDet.Pytorch

Modified from bioharn

from liberator.closer import Closer

closer = Closer()
closer.add_static('EfficientDet', 'models/efficientdet.py')
closer.expand(['models'])
print(closer.current_sourcecode())
"""
import kwimage
import ndsampler
import ubelt as ub
from delayed_image.channel_spec import ChannelSpec
import collections
import re
from torch.utils import model_zoo
import math
import torch
from functools import partial
import warnings
import torch.nn.functional as F
import numpy as np
import torch.nn as nn
from geowatch.utils.util_netharn import InputNorm
from geowatch.utils.util_netharn import padded_collate


def normal_init(module, mean=0, std=1, bias=0):
    nn.init.normal_(module.weight, mean, std)
    if hasattr(module, 'bias'):
        nn.init.constant_(module.bias, bias)


def bias_init_with_prob(prior_prob):
    """ initialize conv/fc bias value according to giving probablity"""
    bias_init = float(-np.log((1 - prior_prob) / prior_prob))
    return bias_init


def conv_ws_2d(input,
               weight,
               bias=None,
               stride=1,
               padding=0,
               dilation=1,
               groups=1,
               eps=1e-5):
    c_in = weight.size(0)
    weight_flat = weight.view(c_in, -1)
    mean = weight_flat.mean(dim=1, keepdim=True).view(c_in, 1, 1, 1)
    std = weight_flat.std(dim=1, keepdim=True).view(c_in, 1, 1, 1)
    weight = (weight - mean) / (std + eps)
    return F.conv2d(input, weight, bias, stride, padding, dilation, groups)


class ConvWS2d(nn.Conv2d):
    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 stride=1,
                 padding=0,
                 dilation=1,
                 groups=1,
                 bias=True,
                 eps=1e-5):
        super(ConvWS2d, self).__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias)
        self.eps = eps

    def forward(self, x):
        return conv_ws_2d(x, self.weight, self.bias, self.stride, self.padding,
                          self.dilation, self.groups, self.eps)


norm_cfg = {
    'BN': (
        'bn', nn.BatchNorm2d), 'SyncBN': (
            'bn', nn.SyncBatchNorm), 'GN': (
                'gn', nn.GroupNorm)}


conv_cfg = {'Conv': nn.Conv2d, 'ConvWS': ConvWS2d}


def build_norm_layer(cfg, num_features, postfix=''):
    """ Build normalization layer
    Args:
        cfg (dict): cfg should contain:
            type (str): identify norm layer type.
            layer args: args needed to instantiate a norm layer.
            requires_grad (bool): [optional] whether stop gradient updates
        num_features (int): number of channels from input.
        postfix (int, str): appended into norm abbreviation to
            create named layer.
    Returns:
        name (str): abbreviation + postfix
        layer (nn.Module): created norm layer
    """
    assert isinstance(cfg, dict) and 'type' in cfg
    cfg_ = cfg.copy()

    layer_type = cfg_.pop('type')
    if layer_type not in norm_cfg:
        raise KeyError('Unrecognized norm type {}'.format(layer_type))
    else:
        abbr, norm_layer = norm_cfg[layer_type]
        if norm_layer is None:
            raise NotImplementedError

    assert isinstance(postfix, (int, str))
    name = abbr + str(postfix)

    requires_grad = cfg_.pop('requires_grad', True)
    cfg_.setdefault('eps', 1e-5)
    if layer_type != 'GN':
        layer = norm_layer(num_features, **cfg_)
        if layer_type == 'SyncBN':
            layer._specify_ddp_gpu_num(1)
    else:
        assert 'num_groups' in cfg_
        layer = norm_layer(num_channels=num_features, **cfg_)

    for param in layer.parameters():
        param.requires_grad = requires_grad

    return name, layer


def build_conv_layer(cfg, *args, **kwargs):
    """ Build convolution layer
    Args:
        cfg (None or dict): cfg should contain:
            type (str): identify conv layer type.
            layer args: args needed to instantiate a conv layer.
    Returns:
        layer (nn.Module): created conv layer
    """
    if cfg is None:
        cfg_ = dict(type='Conv')
    else:
        assert isinstance(cfg, dict) and 'type' in cfg
        cfg_ = cfg.copy()

    layer_type = cfg_.pop('type')
    if layer_type not in conv_cfg:
        raise KeyError('Unrecognized norm type {}'.format(layer_type))
    else:
        conv_layer = conv_cfg[layer_type]

    layer = conv_layer(*args, **kwargs, **cfg_)

    return layer


class ConvModule(nn.Module):
    """A conv block that contains conv/norm/activation layers.
    Args:
        in_channels (int): Same as nn.Conv2d.
        out_channels (int): Same as nn.Conv2d.
        kernel_size (int or tuple[int]): Same as nn.Conv2d.
        stride (int or tuple[int]): Same as nn.Conv2d.
        padding (int or tuple[int]): Same as nn.Conv2d.
        dilation (int or tuple[int]): Same as nn.Conv2d.
        groups (int): Same as nn.Conv2d.
        bias (bool or str): If specified as `auto`, it will be decided by the
            norm_cfg. Bias will be set as True if norm_cfg is None, otherwise
            False.
        conv_cfg (dict): Config dict for convolution layer.
        norm_cfg (dict): Config dict for normalization layer.
        activation (str or None): Activation type, "ReLU" by default.
        inplace (bool): Whether to use inplace mode for activation.
        order (tuple[str]): The order of conv/norm/activation layers. It is a
            sequence of "conv", "norm" and "act". Examples are
            ("conv", "norm", "act") and ("act", "conv", "norm").
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 stride=1,
                 padding=0,
                 dilation=1,
                 groups=1,
                 bias='auto',
                 conv_cfg=None,
                 norm_cfg=None,
                 activation='relu',
                 inplace=True,
                 order=('conv', 'norm', 'act')):
        super(ConvModule, self).__init__()
        assert conv_cfg is None or isinstance(conv_cfg, dict)
        assert norm_cfg is None or isinstance(norm_cfg, dict)
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.activation = activation
        self.inplace = inplace
        self.order = order
        assert isinstance(self.order, tuple) and len(self.order) == 3
        assert set(order) == set(['conv', 'norm', 'act'])

        self.with_norm = norm_cfg is not None
        self.with_activatation = activation is not None
        # if the conv layer is before a norm layer, bias is unnecessary.
        if bias == 'auto':
            bias = False if self.with_norm else True
        self.with_bias = bias

        if self.with_norm and self.with_bias:
            warnings.warn('ConvModule has norm and bias at the same time')

        # build convolution layer
        self.conv = build_conv_layer(
            conv_cfg,
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias)
        # export the attributes of self.conv to a higher level for convenience
        self.in_channels = self.conv.in_channels
        self.out_channels = self.conv.out_channels
        self.kernel_size = self.conv.kernel_size
        self.stride = self.conv.stride
        self.padding = self.conv.padding
        self.dilation = self.conv.dilation
        self.transposed = self.conv.transposed
        self.output_padding = self.conv.output_padding
        self.groups = self.conv.groups

        # build normalization layers
        if self.with_norm:
            # norm layer is after conv layer
            if order.index('norm') > order.index('conv'):
                norm_channels = out_channels
            else:
                norm_channels = in_channels
            self.norm_name, norm = build_norm_layer(norm_cfg, norm_channels)
            self.add_module(self.norm_name, norm)

        # build activation layer
        if self.with_activatation:
            # TODO: introduce `act_cfg` and supports more activation layers
            if self.activation not in ['relu']:
                raise ValueError('{} is currently not supported.'.format(
                    self.activation))
            if self.activation == 'relu':
                self.activate = nn.ReLU(inplace=inplace)

    @property
    def norm(self):
        return getattr(self, self.norm_name)

    def forward(self, x, activate=True, norm=True):
        for layer in self.order:
            if layer == 'conv':
                x = self.conv(x)
            elif layer == 'norm' and norm and self.with_norm:
                x = self.norm(x)
            elif layer == 'act' and activate and self.with_activatation:
                x = self.activate(x)
        return x


def multi_apply(func, *args, **kwargs):
    pfunc = partial(func, **kwargs) if kwargs else func
    map_results = map(pfunc, *args)
    return tuple(map(list, zip(*map_results)))


class RetinaHead(nn.Module):
    """
    An anchor-based head used in [1]_.
    The head contains two subnetworks. The first classifies anchor boxes and
    the second regresses deltas for the anchors.
    References:
        .. [1]  https://arxiv.org/pdf/1708.02002.pdf

    Example:
        >>> from geowatch.tasks.fusion.methods.efficientdet import RetinaHead
        >>> import torch
        >>> self = RetinaHead(11, 7)
        >>> x = torch.rand(1, 7, 32, 32)
        >>> cls_score, bbox_pred = self.forward_single(x)
        >>> # Each anchor predicts a score for each class except background
        >>> cls_per_anchor = cls_score.shape[2]
        >>> box_per_anchor = bbox_pred.shape[2]
        >>> assert cls_per_anchor == (self.num_classes)
        >>> assert box_per_anchor == 4
    """

    def __init__(self,
                 num_classes,
                 in_channels,
                 feat_channels=256,
                 stacked_convs=4,
                 num_anchors=9,
                 conv_cfg=None,
                 norm_cfg=None,
                 **kwargs):
        super(RetinaHead, self).__init__()
        self.in_channels = in_channels
        self.num_classes = num_classes
        self.feat_channels = feat_channels
        self.stacked_convs = stacked_convs
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.cls_out_channels = num_classes
        self.num_anchors = num_anchors
        self._init_layers()

    def _init_layers(self):
        self.relu = nn.ReLU(inplace=True)
        self.cls_convs = nn.ModuleList()
        self.reg_convs = nn.ModuleList()
        for i in range(self.stacked_convs):
            chn = self.in_channels if i == 0 else self.feat_channels
            self.cls_convs.append(
                ConvModule(
                    chn,
                    self.feat_channels,
                    3,
                    stride=1,
                    padding=1,
                    conv_cfg=self.conv_cfg,
                    norm_cfg=self.norm_cfg))
            self.reg_convs.append(
                ConvModule(
                    chn,
                    self.feat_channels,
                    3,
                    stride=1,
                    padding=1,
                    conv_cfg=self.conv_cfg,
                    norm_cfg=self.norm_cfg))
        self.retina_cls = nn.Conv2d(
            self.feat_channels,
            self.num_anchors * self.cls_out_channels,
            3,
            padding=1)
        self.retina_reg = nn.Conv2d(
            self.feat_channels, self.num_anchors * 4, 3, padding=1)
        self.output_act = nn.Sigmoid()

    def init_weights(self):
        for m in self.cls_convs:
            normal_init(m.conv, std=0.01)
        for m in self.reg_convs:
            normal_init(m.conv, std=0.01)
        bias_cls = bias_init_with_prob(0.01)
        normal_init(self.retina_cls, std=0.01, bias=bias_cls)
        normal_init(self.retina_reg, std=0.01)

    def forward_single(self, x):
        cls_feat = x
        reg_feat = x
        for cls_conv in self.cls_convs:
            cls_feat = cls_conv(cls_feat)
        for reg_conv in self.reg_convs:
            reg_feat = reg_conv(reg_feat)

        cls_energy = self.retina_cls(cls_feat)
        cls_score = self.output_act(cls_energy)
        # out is B x C x W x H, with C = n_classes + n_anchors
        cls_score = cls_score.permute(0, 2, 3, 1)
        batch_size, width, height, channels = cls_score.shape
        cls_score = cls_score.view(
            batch_size, width, height, self.num_anchors, self.num_classes)
        cls_score = cls_score.contiguous().view(x.size(0), -1, self.num_classes)

        bbox_pred = self.retina_reg(reg_feat)
        bbox_pred = bbox_pred.permute(0, 2, 3, 1)
        bbox_pred = bbox_pred.contiguous().view(bbox_pred.size(0), -1, 4)
        return cls_score, bbox_pred

    def forward(self, feats):
        return multi_apply(self.forward_single, feats)


def calc_iou(a, b):
    area = (b[:, 2] - b[:, 0]) * (b[:, 3] - b[:, 1])

    iw = torch.min(torch.unsqueeze(
        a[:, 2], dim=1), b[:, 2]) - torch.max(torch.unsqueeze(a[:, 0], 1), b[:, 0])
    ih = torch.min(torch.unsqueeze(
        a[:, 3], dim=1), b[:, 3]) - torch.max(torch.unsqueeze(a[:, 1], 1), b[:, 1])

    iw = torch.clamp(iw, min=0)
    ih = torch.clamp(ih, min=0)

    ua = torch.unsqueeze((a[:, 2] - a[:, 0]) *
                         (a[:, 3] - a[:, 1]), dim=1) + area - iw * ih

    ua = torch.clamp(ua, min=1e-8)

    intersection = iw * ih

    IoU = intersection / ua

    return IoU


class FocalLoss(nn.Module):
    """
    TODO:
        validate loss formulation

        original derived from
        https://github.com/kuangliu/pytorch-retinanet/blob/master/loss.py

        Alternate impl
        https://github.com/facebookresearch/detectron2/blob/master/detectron2/modeling/meta_arch/retinanet.py
    """
    # def __init__(self):

    def forward(criterion, classifications, regressions, anchors, annotations, ignore_idxs=None):
        alpha = 0.25
        gamma = 2.0
        batch_size = classifications.shape[0]
        classification_losses = []
        regression_losses = []

        anchor = anchors[0, :, :]

        if 0:
            anchor_boxes = kwimage.Boxes(anchors[0], 'tlbr').numpy()
            anchor_dsizes = list(zip(np.round(anchor_boxes.width.ravel()).tolist(), np.round(anchor_boxes.height.ravel()).tolist()))
            ub.dict_hist(anchor_dsizes)
            np.sqrt(anchor_boxes.area)

        # TODO :
        # [ ] - annotation weights
        # [ ] - ignore boxes
        # [ ] - option to ignore coarser pyramid levels in architecture

        if 0:
            # Equivalent impl but with 1 fewer lines and no magic numbers
            anchor_cxywh = kwimage.Boxes(anchor, 'tlbr').to_cxywh()
            (anchor_ctr_x, anchor_ctr_y,
             anchor_width, anchor_heights) = anchor_cxywh.data.T
        else:
            anchor_widths = anchor[:, 2] - anchor[:, 0]
            anchor_heights = anchor[:, 3] - anchor[:, 1]
            anchor_ctr_x = anchor[:, 0] + 0.5 * anchor_widths
            anchor_ctr_y = anchor[:, 1] + 0.5 * anchor_heights

        device = annotations.device
        for j in range(batch_size):

            classification = classifications[j, :, :]
            regression = regressions[j, :, :]

            bbox_annotation = annotations[j, :, :]
            bbox_annotation = bbox_annotation[bbox_annotation[:, 4] != -1]

            if ignore_idxs:
                ignore_flags = torch.zeros_like(bbox_annotation[:, 4], dtype=torch.bool)
                for idx in ignore_idxs:
                    ignore_flags |= (bbox_annotation[:, 4] == idx)

            if bbox_annotation.view(-1).shape[0] == 0:
                regression_losses.append(torch.tensor(0.0, requires_grad=True, device=device))
                classification_losses.append(torch.tensor(0.0, requires_grad=True, device=device))

                continue

            clf_eps = 1e-4
            classification = torch.clamp(classification, clf_eps, 1.0 - clf_eps)

            # num_anchors x num_annotations
            IoU = calc_iou(anchor, bbox_annotation[:, :4])

            # For each anchor, find its most similar true box
            # As each anchor is assigned to at most one object box
            IoU_max, IoU_argmax = torch.max(IoU, dim=1)  # num_anchors x 1

            #import pdb
            # pdb.set_trace()

            # compute the loss for classification
            targets = torch.full(classification.shape, fill_value=-1,
                                 device=device, dtype=torch.float)

            # anchor_fg_iou_thresh = 0.5  # original
            # anchor_bg_iou_thresh = 0.4  # original
            anchor_fg_iou_thresh = 0.1  # 0.1 did work
            anchor_bg_iou_thresh = 0.01

            # assign anchors that don't overlap well to background
            # assign anchors that overlap well to foreground
            # note: anchors between iou thresholds are ignored
            negative_flags = IoU_max <= anchor_bg_iou_thresh
            positive_flags = IoU_max > anchor_fg_iou_thresh

            targets[negative_flags, :] = 0
            # TODO: if there is a background class set that to 1.

            num_positive_anchors = positive_flags.sum()

            assigned_annotations = bbox_annotation[IoU_argmax, :]

            positive_cidxs = assigned_annotations[positive_flags, 4].long()

            targets[positive_flags, :] = 0
            targets[positive_flags, positive_cidxs] = 1

            alpha_factor = torch.ones(targets.shape).to(device) * alpha

            alpha_factor = torch.where(targets == 1.,
                                       alpha_factor, 1. - alpha_factor)
            focal_weight = torch.where(targets == 1.,
                                       1. - classification, classification)
            focal_weight = alpha_factor * torch.pow(focal_weight, gamma)

            # bce2 = -(targets * torch.log(classification) +
            #         (1.0 - targets) * torch.log(1.0 - classification))
            bce = nn.functional.binary_cross_entropy(
                classification, targets, reduction='none')
            # cls_loss = focal_weight * torch.pow(bce, gamma)
            cls_loss = focal_weight * bce
            # cls_loss = bce

            cls_loss = torch.where(
                targets != -1.0, cls_loss,
                torch.zeros(cls_loss.shape, device=device))

            clf_loss_norm = (
                cls_loss.sum() / torch.clamp(num_positive_anchors.float(), min=1.0)
            )

            classification_losses.append(clf_loss_norm)

            # compute the loss for regression
            if positive_flags.sum() > 0:
                assigned_annotations = assigned_annotations[positive_flags, :]

                anchor_widths_pi = anchor_widths[positive_flags]
                anchor_heights_pi = anchor_heights[positive_flags]
                anchor_ctr_x_pi = anchor_ctr_x[positive_flags]
                anchor_ctr_y_pi = anchor_ctr_y[positive_flags]

                gt_widths = assigned_annotations[:, 2] - assigned_annotations[:, 0]
                gt_heights = assigned_annotations[:, 3] - assigned_annotations[:, 1]
                gt_ctr_x = assigned_annotations[:, 0] + 0.5 * gt_widths
                gt_ctr_y = assigned_annotations[:, 1] + 0.5 * gt_heights

                # clip minimum truth size to 1
                gt_widths = torch.clamp(gt_widths, min=1)
                gt_heights = torch.clamp(gt_heights, min=1)

                targets_dx = (gt_ctr_x - anchor_ctr_x_pi) / anchor_widths_pi
                targets_dy = (gt_ctr_y - anchor_ctr_y_pi) / anchor_heights_pi
                targets_dw = torch.log(gt_widths / anchor_widths_pi)
                targets_dh = torch.log(gt_heights / anchor_heights_pi)

                targets = torch.stack(
                    (targets_dx, targets_dy, targets_dw, targets_dh))
                targets = targets.t()

                # TODO: must be paramatarized via whatever the BBoxTransform.std / mean is
                targets = targets / torch.Tensor([[0.1, 0.1, 0.2, 0.2]]).to(device)

                if 0:
                    # This is the formulation for the actual box positions
                    # but it is unused here
                    tf = BBoxTransform()
                    tf(anchor[None, :], regression[None, :])

                # negative_indices = 1 + (~positive_flags)

                regression_pi = regression[positive_flags, :]
                regression_diff = torch.abs(targets - regression_pi)

                # This is a smooth L1 loss
                regression_loss = torch.where(
                    regression_diff < (1.0 / 9.0),
                    0.5 * 9.0 * torch.pow(regression_diff, 2),
                    regression_diff - 0.5 / 9.0
                )
                regression_losses.append(regression_loss.mean())
            else:
                regression_losses.append(torch.tensor(0).float().to(device))

        clf_loss = torch.stack(classification_losses).mean(dim=0, keepdim=True)
        regression_loss = torch.stack(regression_losses).mean(dim=0, keepdim=True)

        return clf_loss, regression_loss


def drop_connect(inputs, p, training):
    """ Drop connect. """
    if not training:
        return inputs
    batch_size = inputs.shape[0]
    keep_prob = 1 - p
    random_tensor = keep_prob
    random_tensor += torch.rand([batch_size, 1, 1, 1],
                                dtype=inputs.dtype, device=inputs.device)
    binary_tensor = torch.floor(random_tensor)
    output = inputs / keep_prob * binary_tensor
    return output


def round_repeats(repeats, global_params):
    """ Round number of filters based on depth multiplier. """
    multiplier = global_params.depth_coefficient
    if not multiplier:
        return repeats
    return int(math.ceil(multiplier * repeats))


def round_filters(filters, global_params):
    """ Calculate and round number of filters based on depth multiplier. """
    multiplier = global_params.width_coefficient
    if not multiplier:
        return filters
    divisor = global_params.depth_divisor
    min_depth = global_params.min_depth
    filters *= multiplier
    min_depth = min_depth or divisor
    new_filters = max(min_depth, int(
        filters + divisor / 2) // divisor * divisor)
    if new_filters < 0.9 * filters:  # prevent rounding by more than 10%
        new_filters += divisor
    return int(new_filters)


url_map = {
    # Note: these urls no longer work
    'efficientnet-b0': 'http://storage.googleapis.com/public-models/efficientnet/efficientnet-b0-355c32eb.pth',
    'efficientnet-b1': 'http://storage.googleapis.com/public-models/efficientnet/efficientnet-b1-f1951068.pth',
    'efficientnet-b2': 'http://storage.googleapis.com/public-models/efficientnet/efficientnet-b2-8bb594d6.pth',
    'efficientnet-b3': 'http://storage.googleapis.com/public-models/efficientnet/efficientnet-b3-5fb5a3c3.pth',
    'efficientnet-b4': 'http://storage.googleapis.com/public-models/efficientnet/efficientnet-b4-6ed6700e.pth',
    'efficientnet-b5': 'http://storage.googleapis.com/public-models/efficientnet/efficientnet-b5-b6417697.pth',
    'efficientnet-b6': 'http://storage.googleapis.com/public-models/efficientnet/efficientnet-b6-c76e70fd.pth',
    'efficientnet-b7': 'http://storage.googleapis.com/public-models/efficientnet/efficientnet-b7-dcc49843.pth',
}


def load_pretrained_weights(model, model_name, load_fc=True):
    """ Loads pretrained weights, and downloads if loading for the first time. """
    state_dict = model_zoo.load_url(url_map[model_name])

    if load_fc:
        model.load_state_dict(state_dict)
    else:
        state_dict.pop('_fc.weight')
        state_dict.pop('_fc.bias')
        res = model.load_state_dict(state_dict, strict=False)
        assert set(res.missing_keys) == set(
            ['_fc.weight', '_fc.bias']), 'issue loading pretrained weights'
    print('Loaded pretrained weights for {}'.format(model_name))


class Identity(nn.Module):
    def __init__(self, ):
        super(Identity, self).__init__()

    def forward(self, input):
        return input


class Conv2dStaticSamePadding(nn.Conv2d):
    """ 2D Convolutions like TensorFlow, for a fixed image size"""

    def __init__(self, in_channels, out_channels,
                 kernel_size, image_size=None, **kwargs):
        super().__init__(in_channels, out_channels, kernel_size, **kwargs)
        self.stride = self.stride if len(self.stride) == 2 else [
            self.stride[0]] * 2

        # Calculate padding based on image size and save it
        assert image_size is not None
        ih, iw = image_size if isinstance(image_size, list) else [
            image_size, image_size]
        kh, kw = self.weight.size()[-2:]
        sh, sw = self.stride
        oh, ow = math.ceil(ih / sh), math.ceil(iw / sw)
        pad_h = max((oh - 1) * self.stride[0] +
                    (kh - 1) * self.dilation[0] + 1 - ih, 0)
        pad_w = max((ow - 1) * self.stride[1] +
                    (kw - 1) * self.dilation[1] + 1 - iw, 0)
        if pad_h > 0 or pad_w > 0:
            self.static_padding = nn.ZeroPad2d(
                (pad_w // 2, pad_w - pad_w // 2, pad_h // 2, pad_h - pad_h // 2))
        else:
            self.static_padding = Identity()

    def forward(self, x):
        x = self.static_padding(x)
        x = F.conv2d(x, self.weight, self.bias, self.stride,
                     self.padding, self.dilation, self.groups)
        return x


class Conv2dDynamicSamePadding(nn.Conv2d):
    """ 2D Convolutions like TensorFlow, for a dynamic image size """

    def __init__(self, in_channels, out_channels, kernel_size,
                 stride=1, dilation=1, groups=1, bias=True):
        super().__init__(in_channels, out_channels,
                         kernel_size, stride, 0, dilation, groups, bias)
        self.stride = self.stride if len(self.stride) == 2 else [
            self.stride[0]] * 2

    def forward(self, x):
        ih, iw = x.size()[-2:]
        kh, kw = self.weight.size()[-2:]
        sh, sw = self.stride
        oh, ow = math.ceil(ih / sh), math.ceil(iw / sw)
        pad_h = max((oh - 1) * self.stride[0] +
                    (kh - 1) * self.dilation[0] + 1 - ih, 0)
        pad_w = max((ow - 1) * self.stride[1] +
                    (kw - 1) * self.dilation[1] + 1 - iw, 0)
        if pad_h > 0 or pad_w > 0:
            x = F.pad(x, [pad_w // 2, pad_w - pad_w //
                          2, pad_h // 2, pad_h - pad_h // 2])
        return F.conv2d(x, self.weight, self.bias, self.stride,
                        self.padding, self.dilation, self.groups)


def get_same_padding_conv2d(image_size=None):
    """ Chooses static padding if you have specified an image size, and dynamic padding otherwise.
        Static padding is necessary for ONNX exporting of models. """
    if image_size is None:
        return Conv2dDynamicSamePadding
    else:
        return partial(Conv2dStaticSamePadding, image_size=image_size)


BlockArgs = collections.namedtuple('BlockArgs',
                                   ['kernel_size',
                                    'num_repeat',
                                    'input_filters',
                                    'output_filters',
                                    'expand_ratio',
                                    'id_skip',
                                    'stride',
                                    'se_ratio'])


GlobalParams = collections.namedtuple('GlobalParams',
                                      ['batch_norm_momentum',
                                       'batch_norm_epsilon',
                                       'dropout_rate',
                                       'num_classes',
                                       'width_coefficient',
                                       'depth_coefficient',
                                       'depth_divisor',
                                       'min_depth',
                                       'drop_connect_rate',
                                       'image_size'])


class BlockDecoder(object):
    """ Block Decoder for readability, straight from the official TensorFlow repository """

    @staticmethod
    def _decode_block_string(block_string):
        """ Gets a block through a string notation of arguments. """
        assert isinstance(block_string, str)

        ops = block_string.split('_')
        options = {}
        for op in ops:
            splits = re.split(r'(\d.*)', op)
            if len(splits) >= 2:
                key, value = splits[:2]
                options[key] = value

        # Check stride
        assert (('s' in options and len(options['s']) == 1) or
                (len(options['s']) == 2 and options['s'][0] == options['s'][1]))

        return BlockArgs(
            kernel_size=int(options['k']),
            num_repeat=int(options['r']),
            input_filters=int(options['i']),
            output_filters=int(options['o']),
            expand_ratio=int(options['e']),
            id_skip=('noskip' not in block_string),
            se_ratio=float(options['se']) if 'se' in options else None,
            stride=[int(options['s'][0])])

    @staticmethod
    def _encode_block_string(block):
        """Encodes a block to a string."""
        args = [
            'r%d' % block.num_repeat,
            'k%d' % block.kernel_size,
            's%d%d' % (block.strides[0], block.strides[1]),
            'e%s' % block.expand_ratio,
            'i%d' % block.input_filters,
            'o%d' % block.output_filters
        ]
        if 0 < block.se_ratio <= 1:
            args.append('se%s' % block.se_ratio)
        if block.id_skip is False:
            args.append('noskip')
        return '_'.join(args)

    @staticmethod
    def decode(string_list):
        """
        Decodes a list of string notations to specify blocks inside the network.
        :param string_list: a list of strings, each string is a notation of block
        :return: a list of BlockArgs namedtuples of block args
        """
        assert isinstance(string_list, list)
        blocks_args = []
        for block_string in string_list:
            blocks_args.append(BlockDecoder._decode_block_string(block_string))
        return blocks_args

    @staticmethod
    def encode(blocks_args):
        """
        Encodes a list of BlockArgs to a list of strings.
        :param blocks_args: a list of BlockArgs namedtuples of block args
        :return: a list of strings, each string is a notation of block
        """
        block_strings = []
        for block in blocks_args:
            block_strings.append(BlockDecoder._encode_block_string(block))
        return block_strings


def efficientnet_params(model_name):
    """ Map EfficientNet model name to parameter coefficients. """
    params_dict = {
        # Coefficients:   width,depth,res,dropout
        'efficientnet-b0': (1.0, 1.0, 224, 0.2),
        'efficientnet-b1': (1.0, 1.1, 240, 0.2),
        'efficientnet-b2': (1.1, 1.2, 260, 0.3),
        'efficientnet-b3': (1.2, 1.4, 300, 0.3),
        'efficientnet-b4': (1.4, 1.8, 380, 0.4),
        'efficientnet-b5': (1.6, 2.2, 456, 0.4),
        'efficientnet-b6': (1.8, 2.6, 528, 0.5),
        'efficientnet-b7': (2.0, 3.1, 600, 0.5),
    }
    return params_dict[model_name]


def efficientnet(width_coefficient=None, depth_coefficient=None, dropout_rate=0.2,
                 drop_connect_rate=0.2, image_size=None, num_classes=1000):
    """ Creates a efficientnet model. """

    blocks_args = [
        'r1_k3_s11_e1_i32_o16_se0.25', 'r2_k3_s22_e6_i16_o24_se0.25',
        'r2_k5_s22_e6_i24_o40_se0.25', 'r3_k3_s22_e6_i40_o80_se0.25',
        'r3_k5_s22_e6_i80_o112_se0.25', 'r4_k5_s22_e6_i112_o192_se0.25',
        'r1_k3_s22_e6_i192_o320_se0.25',
    ]
    blocks_args = BlockDecoder.decode(blocks_args)

    global_params = GlobalParams(
        batch_norm_momentum=0.99,
        batch_norm_epsilon=1e-3,
        dropout_rate=dropout_rate,
        drop_connect_rate=drop_connect_rate,
        # data_format='channels_last',  # removed, this is always true in
        # PyTorch
        num_classes=num_classes,
        width_coefficient=width_coefficient,
        depth_coefficient=depth_coefficient,
        depth_divisor=8,
        min_depth=None,
        image_size=image_size,
    )

    return blocks_args, global_params


def get_model_params(model_name, override_params):
    """ Get the block args and global params for a given model """
    if model_name.startswith('efficientnet'):
        w, d, s, p = efficientnet_params(model_name)
        # note: all models have drop connect rate = 0.2
        blocks_args, global_params = efficientnet(
            width_coefficient=w, depth_coefficient=d, dropout_rate=p, image_size=s)
    else:
        raise NotImplementedError(
            'model name is not pre-defined: %s' % model_name)
    if override_params:
        # ValueError will be raised here if override_params has fields not
        # included in global_params.
        global_params = global_params._replace(**override_params)
    return blocks_args, global_params


class Swish(nn.Module):
    def forward(self, x):
        return x * torch.sigmoid(x)


class SwishImplementation(torch.autograd.Function):
    @staticmethod
    def forward(ctx, i):
        result = i * torch.sigmoid(i)
        ctx.save_for_backward(i)
        return result

    @staticmethod
    def backward(ctx, grad_output):
        i = ctx.saved_variables[0]
        sigmoid_i = torch.sigmoid(i)
        return grad_output * (sigmoid_i * (1 + i * (1 - sigmoid_i)))


class MemoryEfficientSwish(nn.Module):
    def forward(self, x):
        return SwishImplementation.apply(x)


class MBConvBlock(nn.Module):
    """
    Mobile Inverted Residual Bottleneck Block
    Args:
        block_args (namedtuple): BlockArgs, see above
        global_params (namedtuple): GlobalParam, see above
    Attributes:
        has_se (bool): Whether the block contains a Squeeze and Excitation layer.
    """

    def __init__(self, block_args, global_params):
        super().__init__()
        self._block_args = block_args
        self._bn_mom = 1 - global_params.batch_norm_momentum
        self._bn_eps = global_params.batch_norm_epsilon
        self.has_se = (self._block_args.se_ratio is not None) and (
            0 < self._block_args.se_ratio <= 1)
        self.id_skip = block_args.id_skip  # skip connection and drop connect

        # Get static or dynamic convolution depending on image size
        Conv2d = get_same_padding_conv2d(image_size=global_params.image_size)

        # Expansion phase
        inp = self._block_args.input_filters  # number of input channels
        oup = self._block_args.input_filters * \
            self._block_args.expand_ratio  # number of output channels
        if self._block_args.expand_ratio != 1:
            self._expand_conv = Conv2d(
                in_channels=inp, out_channels=oup, kernel_size=1, bias=False)
            self._bn0 = nn.BatchNorm2d(
                num_features=oup, momentum=self._bn_mom, eps=self._bn_eps)
        # Depthwise convolution phase
        k = self._block_args.kernel_size
        s = self._block_args.stride
        self._depthwise_conv = Conv2d(
            in_channels=oup, out_channels=oup, groups=oup,  # groups makes it depthwise
            kernel_size=k, stride=s, bias=False)
        self._bn1 = nn.BatchNorm2d(
            num_features=oup, momentum=self._bn_mom, eps=self._bn_eps)

        # Squeeze and Excitation layer, if desired
        if self.has_se:
            num_squeezed_channels = max(
                1, int(self._block_args.input_filters * self._block_args.se_ratio))
            self._se_reduce = Conv2d(
                in_channels=oup, out_channels=num_squeezed_channels, kernel_size=1)
            self._se_expand = Conv2d(
                in_channels=num_squeezed_channels, out_channels=oup, kernel_size=1)

        # Output phase
        final_oup = self._block_args.output_filters
        self._project_conv = Conv2d(
            in_channels=oup, out_channels=final_oup, kernel_size=1, bias=False)
        self._bn2 = nn.BatchNorm2d(
            num_features=final_oup, momentum=self._bn_mom, eps=self._bn_eps)
        self._swish = MemoryEfficientSwish()

    def forward(self, inputs, drop_connect_rate=None):
        """
        :param inputs: input tensor
        :param drop_connect_rate: drop connect rate (float, between 0 and 1)
        :return: output of block
        """

        # Expansion and Depthwise Convolution
        x = inputs
        if self._block_args.expand_ratio != 1:
            x = self._swish(self._bn0(self._expand_conv(inputs)))

        x = self._swish(self._bn1(self._depthwise_conv(x)))

        # Squeeze and Excitation
        if self.has_se:
            x_squeezed = F.adaptive_avg_pool2d(x, 1)
            x_squeezed = self._se_expand(
                self._swish(self._se_reduce(x_squeezed)))
            x = torch.sigmoid(x_squeezed) * x

        x = self._bn2(self._project_conv(x))

        # Skip connection and drop connect
        input_filters, output_filters = self._block_args.input_filters, self._block_args.output_filters
        if self.id_skip and self._block_args.stride == 1 and input_filters == output_filters:
            if drop_connect_rate:
                x = drop_connect(x, p=drop_connect_rate,
                                 training=self.training)
            x = x + inputs  # skip connection
        return x

    def set_swish(self, memory_efficient=True):
        """Sets swish function as memory efficient (for training) or standard (for export)"""
        self._swish = MemoryEfficientSwish() if memory_efficient else Swish()


class EfficientNet(nn.Module):
    """
    An EfficientNet model. Most easily loaded with the .from_name or .from_pretrained methods
    Args:
        blocks_args (list): A list of BlockArgs to construct blocks
        global_params (namedtuple): A set of GlobalParams shared between blocks
    Example:
        model = EfficientNet.from_pretrained('efficientnet-b0')
    """

    def __init__(self, blocks_args=None, global_params=None):
        super().__init__()
        assert isinstance(blocks_args, list), 'blocks_args should be a list'
        assert len(blocks_args) > 0, 'block args must be greater than 0'
        self._global_params = global_params
        self._blocks_args = blocks_args

        # Get static or dynamic convolution depending on image size
        Conv2d = get_same_padding_conv2d(image_size=global_params.image_size)

        # Batch norm parameters
        bn_mom = 1 - self._global_params.batch_norm_momentum
        bn_eps = self._global_params.batch_norm_epsilon

        # Stem
        in_channels = 3  # rgb
        # number of output channels
        out_channels = round_filters(32, self._global_params)
        self._conv_stem = Conv2d(
            in_channels, out_channels, kernel_size=3, stride=2, bias=False)
        self._bn0 = nn.BatchNorm2d(
            num_features=out_channels, momentum=bn_mom, eps=bn_eps)

        # Build blocks
        self._blocks = nn.ModuleList([])
        for i in range(len(self._blocks_args)):
            # Update block input and output filters based on depth multiplier.
            self._blocks_args[i] = self._blocks_args[i]._replace(
                input_filters=round_filters(
                    self._blocks_args[i].input_filters, self._global_params),
                output_filters=round_filters(
                    self._blocks_args[i].output_filters, self._global_params),
                num_repeat=round_repeats(
                    self._blocks_args[i].num_repeat, self._global_params)
            )

            # The first block needs to take care of stride and filter size
            # increase.
            self._blocks.append(MBConvBlock(
                self._blocks_args[i], self._global_params))
            if self._blocks_args[i].num_repeat > 1:
                self._blocks_args[i] = self._blocks_args[i]._replace(
                    input_filters=self._blocks_args[i].output_filters, stride=1)
            for _ in range(self._blocks_args[i].num_repeat - 1):
                self._blocks.append(MBConvBlock(
                    self._blocks_args[i], self._global_params))

        # Head'efficientdet-d0': 'efficientnet-b0',
        # output of final block
        in_channels = self._blocks_args[len(
            self._blocks_args) - 1].output_filters
        out_channels = round_filters(1280, self._global_params)
        self._conv_head = Conv2d(
            in_channels, out_channels, kernel_size=1, bias=False)
        self._bn1 = nn.BatchNorm2d(
            num_features=out_channels, momentum=bn_mom, eps=bn_eps)

        # Final linear layer
        self._avg_pooling = nn.AdaptiveAvgPool2d(1)
        self._dropout = nn.Dropout(self._global_params.dropout_rate)
        self._fc = nn.Linear(out_channels, self._global_params.num_classes)
        self._swish = MemoryEfficientSwish()

    def set_swish(self, memory_efficient=True):
        """Sets swish function as memory efficient (for training) or standard (for export)"""
        self._swish = MemoryEfficientSwish() if memory_efficient else Swish()
        for block in self._blocks:
            block.set_swish(memory_efficient)

    def extract_features(self, inputs):
        """ Returns output of the final convolution layer """
        # Stem
        x = self._swish(self._bn0(self._conv_stem(inputs)))

        P = []
        index = 0
        num_repeat = 0
        # Blocks
        for idx, block in enumerate(self._blocks):
            drop_connect_rate = self._global_params.drop_connect_rate
            if drop_connect_rate:
                drop_connect_rate *= float(idx) / len(self._blocks)
            x = block(x, drop_connect_rate=drop_connect_rate)
            num_repeat = num_repeat + 1
            if num_repeat == self._blocks_args[index].num_repeat:
                num_repeat = 0
                index = index + 1
                P.append(x)
        return P

    def forward(self, inputs):
        """ Calls extract_features to extract features, applies final linear layer, and returns logits. """
        # Convolution layers
        P = self.extract_features(inputs)
        return P

    @classmethod
    def from_name(cls, model_name, override_params=None):
        cls._check_model_name_is_valid(model_name)
        blocks_args, global_params = get_model_params(
            model_name, override_params)
        return cls(blocks_args, global_params)

    @classmethod
    def from_pretrained(cls, model_name, num_classes=1000, in_channels=3):
        model = cls.from_name(model_name, override_params={
                              'num_classes': num_classes})
        try:
            load_pretrained_weights(
                model, model_name, load_fc=(num_classes == 1000))
        except Exception as ex:
            import warnings
            warnings.warn('Failed to get pretrained weights ex = {!r}'.format(ex))
        if in_channels != 3:
            Conv2d = get_same_padding_conv2d(
                image_size=model._global_params.image_size)
            out_channels = round_filters(32, model._global_params)
            model._conv_stem = Conv2d(
                in_channels, out_channels, kernel_size=3, stride=2, bias=False)
        return model

    # @classmethod
    # def from_pretrained(cls, model_name, num_classes=1000):
    #     model = cls.from_name(model_name, override_params={
    #                           'num_classes': num_classes})
    #     load_pretrained_weights(
    #         model, model_name, load_fc=(num_classes == 1000))

    #     return model

    @classmethod
    def get_image_size(cls, model_name):
        cls._check_model_name_is_valid(model_name)
        _, _, res, _ = efficientnet_params(model_name)
        return res

    @classmethod
    def _check_model_name_is_valid(
            cls, model_name, also_need_pretrained_weights=False):
        """ Validates model name. None that pretrained weights are only available for
        the first four models (efficientnet-b{i} for i in 0,1,2,3) at the moment. """
        num_models = 4 if also_need_pretrained_weights else 8
        valid_models = ['efficientnet-b' + str(i) for i in range(num_models)]
        if model_name not in valid_models:
            raise ValueError('model_name should be one of: ' +
                             ', '.join(valid_models))

    def get_list_features(self):
        list_feature = []
        for idx in range(len(self._blocks_args)):
            list_feature.append(self._blocks_args[idx].output_filters)

        return list_feature


class ClipBoxes(nn.Module):

    def __init__(self, width=None, height=None):
        super(ClipBoxes, self).__init__()

    def forward(self, boxes, img):

        batch_size, num_channels, height, width = img.shape

        boxes[:, :, 0] = torch.clamp(boxes[:, :, 0], min=0)
        boxes[:, :, 1] = torch.clamp(boxes[:, :, 1], min=0)

        boxes[:, :, 2] = torch.clamp(boxes[:, :, 2], max=width)
        boxes[:, :, 3] = torch.clamp(boxes[:, :, 3], max=height)

        return boxes


def xavier_init(module, gain=1, bias=0, distribution='normal'):
    assert distribution in ['uniform', 'normal']
    if distribution == 'uniform':
        nn.init.xavier_uniform_(module.weight, gain=gain)
    else:
        nn.init.xavier_normal_(module.weight, gain=gain)
    if hasattr(module, 'bias'):
        nn.init.constant_(module.bias, bias)


class BiFPNModule(nn.Module):
    def __init__(self,
                 channels,
                 levels,
                 init=0.5,
                 conv_cfg=None,
                 norm_cfg=None,
                 activation=None,
                 eps=0.0001):
        super(BiFPNModule, self).__init__()
        self.activation = activation
        self.eps = eps
        self.levels = levels
        self.bifpn_convs = nn.ModuleList()
        # weighted
        self.w1 = nn.Parameter(torch.Tensor(2, levels).fill_(init))
        self.relu1 = nn.ReLU()
        self.w2 = nn.Parameter(torch.Tensor(3, levels - 2).fill_(init))
        self.relu2 = nn.ReLU()
        for jj in range(2):
            for i in range(self.levels - 1):  # 1,2,3
                fpn_conv = nn.Sequential(
                    ConvModule(
                        channels,
                        channels,
                        3,
                        padding=1,
                        conv_cfg=conv_cfg,
                        norm_cfg=norm_cfg,
                        activation=self.activation,
                        inplace=False)
                )
                self.bifpn_convs.append(fpn_conv)

    # default init_weights for conv(msra) and norm in ConvModule
    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                xavier_init(m, distribution='uniform')

    def forward(self, inputs):
        assert len(inputs) == self.levels
        # build top-down and down-top path with stack
        levels = self.levels
        # w relu
        w1 = self.relu1(self.w1)
        w1 /= torch.sum(w1, dim=0) + self.eps  # normalize
        w2 = self.relu2(self.w2)
        w2 /= torch.sum(w2, dim=0) + self.eps  # normalize
        # build top-down
        idx_bifpn = 0
        pathtd = inputs
        inputs_clone = []
        for in_tensor in inputs:
            inputs_clone.append(in_tensor.clone())

        for i in range(levels - 1, 0, -1):
            pathtd[i - 1] = (w1[0, i - 1] * pathtd[i - 1] + w1[1, i - 1] * F.interpolate(
                pathtd[i], scale_factor=2, mode='nearest')) / (w1[0, i - 1] + w1[1, i - 1] + self.eps)
            pathtd[i - 1] = self.bifpn_convs[idx_bifpn](pathtd[i - 1])
            idx_bifpn = idx_bifpn + 1
        # build down-top
        for i in range(0, levels - 2, 1):
            pathtd[i + 1] = (w2[0, i] * pathtd[i + 1] + w2[1, i] * F.max_pool2d(pathtd[i], kernel_size=2) +
                             w2[2, i] * inputs_clone[i + 1]) / (w2[0, i] + w2[1, i] + w2[2, i] + self.eps)
            pathtd[i + 1] = self.bifpn_convs[idx_bifpn](pathtd[i + 1])
            idx_bifpn = idx_bifpn + 1

        pathtd[levels - 1] = (w1[0, levels - 1] * pathtd[levels - 1] + w1[1, levels - 1] * F.max_pool2d(
            pathtd[levels - 2], kernel_size=2)) / (w1[0, levels - 1] + w1[1, levels - 1] + self.eps)
        pathtd[levels - 1] = self.bifpn_convs[idx_bifpn](pathtd[levels - 1])
        return pathtd


class BIFPN(nn.Module):
    """
    I think this means bidirectional feature pyramid network
    """
    def __init__(self,
                 in_channels,
                 out_channels,
                 num_outs,
                 start_level=0,
                 end_level=-1,
                 stack=1,
                 add_extra_convs=False,
                 extra_convs_on_inputs=True,
                 relu_before_extra_convs=False,
                 no_norm_on_lateral=False,
                 conv_cfg=None,
                 norm_cfg=None,
                 activation=None):
        super(BIFPN, self).__init__()
        assert isinstance(in_channels, list)
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.num_ins = len(in_channels)
        self.num_outs = num_outs
        self.activation = activation
        self.relu_before_extra_convs = relu_before_extra_convs
        self.no_norm_on_lateral = no_norm_on_lateral
        self.stack = stack

        if end_level == -1:
            self.backbone_end_level = self.num_ins
            assert num_outs >= self.num_ins - start_level
        else:
            # if end_level < inputs, no extra level is allowed
            self.backbone_end_level = end_level
            assert end_level <= len(in_channels)
            assert num_outs == end_level - start_level
        self.start_level = start_level
        self.end_level = end_level
        self.add_extra_convs = add_extra_convs
        self.extra_convs_on_inputs = extra_convs_on_inputs

        self.lateral_convs = nn.ModuleList()
        self.fpn_convs = nn.ModuleList()
        self.stack_bifpn_convs = nn.ModuleList()

        for i in range(self.start_level, self.backbone_end_level):
            l_conv = ConvModule(
                in_channels[i],
                out_channels,
                1,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg if not self.no_norm_on_lateral else None,
                activation=self.activation,
                inplace=False)
            self.lateral_convs.append(l_conv)

        for ii in range(stack):
            self.stack_bifpn_convs.append(BiFPNModule(channels=out_channels,
                                                      levels=self.backbone_end_level - self.start_level,
                                                      conv_cfg=conv_cfg,
                                                      norm_cfg=norm_cfg,
                                                      activation=activation))
        # add extra conv layers (e.g., RetinaNet)
        extra_levels = num_outs - self.backbone_end_level + self.start_level
        if add_extra_convs and extra_levels >= 1:
            for i in range(extra_levels):
                if i == 0 and self.extra_convs_on_inputs:
                    in_channels = self.in_channels[self.backbone_end_level - 1]
                else:
                    in_channels = out_channels
                extra_fpn_conv = ConvModule(
                    in_channels,
                    out_channels,
                    3,
                    stride=2,
                    padding=1,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    activation=self.activation,
                    inplace=False)
                self.fpn_convs.append(extra_fpn_conv)
        self.init_weights()

    # default init_weights for conv(msra) and norm in ConvModule
    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                xavier_init(m, distribution='uniform')

    def forward(self, inputs):
        assert len(inputs) == len(self.in_channels)

        # build laterals
        laterals = [
            lateral_conv(inputs[i + self.start_level])
            for i, lateral_conv in enumerate(self.lateral_convs)
        ]

        # part 1: build top-down and down-top path with stack
        used_backbone_levels = len(laterals)
        for bifpn_module in self.stack_bifpn_convs:
            laterals = bifpn_module(laterals)
        outs = laterals
        # part 2: add extra levels
        if self.num_outs > len(outs):
            # use max pool to get more levels on top of outputs
            # (e.g., Faster R-CNN, Mask R-CNN)
            if not self.add_extra_convs:
                for i in range(self.num_outs - used_backbone_levels):
                    outs.append(F.max_pool2d(outs[-1], 1, stride=2))
            # add conv layers on top of original feature maps (RetinaNet)
            else:
                if self.extra_convs_on_inputs:
                    orig = inputs[self.backbone_end_level - 1]
                    outs.append(self.fpn_convs[0](orig))
                else:
                    outs.append(self.fpn_convs[0](outs[-1]))
                for i in range(1, self.num_outs - used_backbone_levels):
                    if self.relu_before_extra_convs:
                        outs.append(self.fpn_convs[i](F.relu(outs[-1])))
                    else:
                        outs.append(self.fpn_convs[i](outs[-1]))
        return tuple(outs)


class BBoxTransform(nn.Module):

    def __init__(self, mean=None, std=None):
        super(BBoxTransform, self).__init__()
        if mean is None:
            self.mean = torch.from_numpy(
                np.array([0, 0, 0, 0]).astype(np.float32))
        else:
            self.mean = mean
        if std is None:
            self.std = torch.from_numpy(
                np.array([0.1, 0.1, 0.2, 0.2]).astype(np.float32))
        else:
            self.std = std

    def forward(self, anchors, regressions):
        """
        boxes = anchors
        deltas = regressions
        """

        widths = anchors[:, :, 2] - anchors[:, :, 0]
        heights = anchors[:, :, 3] - anchors[:, :, 1]
        ctr_x = anchors[:, :, 0] + 0.5 * widths
        ctr_y = anchors[:, :, 1] + 0.5 * heights

        dx = regressions[:, :, 0] * self.std[0] + self.mean[0]
        dy = regressions[:, :, 1] * self.std[1] + self.mean[1]
        dw = regressions[:, :, 2] * self.std[2] + self.mean[2]
        dh = regressions[:, :, 3] * self.std[3] + self.mean[3]

        pred_ctr_x = ctr_x + dx * widths
        pred_ctr_y = ctr_y + dy * heights
        pred_w = torch.exp(dw) * widths
        pred_h = torch.exp(dh) * heights

        pred_boxes_x1 = pred_ctr_x - 0.5 * pred_w
        pred_boxes_y1 = pred_ctr_y - 0.5 * pred_h
        pred_boxes_x2 = pred_ctr_x + 0.5 * pred_w
        pred_boxes_y2 = pred_ctr_y + 0.5 * pred_h

        pred_boxes = torch.stack(
            [pred_boxes_x1, pred_boxes_y1, pred_boxes_x2, pred_boxes_y2], dim=2)

        return pred_boxes


def shift(shape, stride, anchors):
    shift_x = (np.arange(0, shape[1]) + 0.5) * stride
    shift_y = (np.arange(0, shape[0]) + 0.5) * stride

    shift_x, shift_y = np.meshgrid(shift_x, shift_y)

    shifts = np.vstack((
        shift_x.ravel(), shift_y.ravel(),
        shift_x.ravel(), shift_y.ravel()
    )).transpose()

    # add A anchors (1, A, 4) to
    # cell K shifts (K, 1, 4) to get
    # shift anchors (K, A, 4)
    # reshape to (K*A, 4) shifted anchors
    A = anchors.shape[0]
    K = shifts.shape[0]
    all_anchors = (anchors.reshape((1, A, 4)) +
                   shifts.reshape((1, K, 4)).transpose((1, 0, 2)))
    all_anchors = all_anchors.reshape((K * A, 4))

    return all_anchors


def generate_anchors(base_size=16, ratios=None, scales=None):
    """
    Generate anchor (reference) windows by enumerating aspect ratios X
    scales w.r.t. a reference window.
    """

    if ratios is None:
        ratios = np.array([0.5, 1, 2])

    if scales is None:
        scales = np.array([2 ** 0, 2 ** (1.0 / 3.0), 2 ** (2.0 / 3.0)])

    num_anchors = len(ratios) * len(scales)

    # initialize output anchors
    anchors = np.zeros((num_anchors, 4))

    # scale base_size
    anchors[:, 2:] = base_size * np.tile(scales, (2, len(ratios))).T

    # compute areas of anchors
    areas = anchors[:, 2] * anchors[:, 3]

    # correct for ratios
    anchors[:, 2] = np.sqrt(areas / np.repeat(ratios, len(scales)))
    anchors[:, 3] = anchors[:, 2] * np.repeat(ratios, len(scales))

    # transform from (x_ctr, y_ctr, w, h) -> (x1, y1, x2, y2)
    anchors[:, 0::2] -= np.tile(anchors[:, 2] * 0.5, (2, 1)).T
    anchors[:, 1::2] -= np.tile(anchors[:, 3] * 0.5, (2, 1)).T

    return anchors


class Anchors(nn.Module):
    """
    Example:
        >>> self = anchors = Anchors()
        >>> image_shape = (130, 130)
        >>> anchors = self.forward(image_shape)
        >>> anchors = kwimage.Boxes(anchors, 'tlbr')
        >>> print(anchors.to_cxywh())

    """
    def __init__(self, pyramid_levels=None, strides=None,
                 sizes=None, ratios=None, scales=None):
        super(Anchors, self).__init__()

        if pyramid_levels is None:
            self.pyramid_levels = [3, 4, 5, 6, 7]
        else:
            self.pyramid_levels = pyramid_levels
        if strides is None:
            self.strides = [2 ** x for x in self.pyramid_levels]
        if sizes is None:
            self.sizes = [2 ** (x + 2) for x in self.pyramid_levels]
        if ratios is None:
            self.ratios = np.array([0.5, 1, 2])
        if scales is None:
            self.scales = np.array(
                [2 ** 0, 2 ** (1.0 / 3.0), 2 ** (2.0 / 3.0)])

        self.num_per_cell = len(self.scales) * len(self.ratios)

    def forward(self, image_shape, device=None):
        # image_shape = image.shape[2:]
        image_shape = np.array(image_shape)
        level_shapes = [(image_shape + 2 ** x - 1) // (2 ** x)
                        for x in self.pyramid_levels]

        # compute anchors over all pyramid levels
        all_anchors = np.zeros((0, 4)).astype(np.float32)

        for idx, p in enumerate(self.pyramid_levels):
            anchors = generate_anchors(
                base_size=self.sizes[idx],
                ratios=self.ratios,
                scales=self.scales
            )
            level_shape = level_shapes[idx]
            # print('level_shape = {!r}'.format(level_shape))
            shifted_anchors = shift(level_shape, self.strides[idx], anchors)
            # print('shifted_anchors.shape = {!r}'.format(shifted_anchors.shape))
            all_anchors = np.append(all_anchors, shifted_anchors, axis=0)

        all_anchors = np.expand_dims(all_anchors, axis=0)

        return torch.from_numpy(
            all_anchors.astype(np.float32)).to(device)


MODEL_MAP = {
    'efficientdet-d0': 'efficientnet-b0',
    'efficientdet-d1': 'efficientnet-b1',
    'efficientdet-d2': 'efficientnet-b2',
    'efficientdet-d3': 'efficientnet-b3',
    'efficientdet-d4': 'efficientnet-b4',
    'efficientdet-d5': 'efficientnet-b5',
    'efficientdet-d6': 'efficientnet-b6',
    'efficientdet-d7': 'efficientnet-b6',
}


class EfficientDetCoder(object):
    """
    Transforms output of EfficientDet into kwimage.Detections
    """
    def __init__(coder, classes, threshold, iou_threshold):
        coder.classes = classes
        coder.threshold = threshold
        coder.iou_threshold = iou_threshold

    def decode_batch(coder, outputs):
        classifications = outputs['classifications']
        transformed_anchors = outputs['transformed_anchors']

        batch_size = classifications.shape[0]

        n_top = 3  # always return at least 3 boxes
        batch_dets = []
        for bx in range(batch_size):
            pred_probs = classifications[bx]
            pred_score, pred_cidx = pred_probs.max(dim=1)
            pred_coords = transformed_anchors[bx]
            det = kwimage.Detections(
                boxes=kwimage.Boxes(pred_coords, 'tlbr'),
                scores=pred_score,
                probs=pred_probs,
                class_idxs=pred_cidx,
                classes=coder.classes
            )
            flags = det.scores > coder.threshold
            top_idxs = pred_score.argsort()[-n_top:]
            flags[top_idxs] = True
            det = det.compress(flags)
            det = det.non_max_supress(thresh=coder.iou_threshold)
            batch_dets.append(det)
        return batch_dets


class EfficientDet(nn.Module):
    """
    Ignore:
        >>> from geowatch.tasks.fusion.methods.efficientdet import *  # NOQA
        >>> classes = ['class_{:0d}'.format(i) for i in range(81)]
        >>> channels = ChannelSpec.coerce('rgb')
        >>> self = EfficientDet(classes=classes, channels='rgb')
    """

    __BUILTIN_CRITERION__ = True

    def __init__(self, classes=None, input_stats=None, channels=None,
                 network='efficientdet-d0', D_bifpn=3, W_bifpn=88, D_class=3,
                 threshold=0.01, iou_threshold=0.5, n_scales=5):
        super(EfficientDet, self).__init__()

        classes = ndsampler.CategoryTree.coerce(classes)
        self.classes = classes
        num_classes = len(classes)

        if input_stats is None:
            input_stats = {}

        # TODO: use channels when input is not RGB
        self.channels = ChannelSpec.coerce(channels)
        chan_keys = list(self.channels.keys())
        if len(chan_keys) != 1:
            raise ValueError('this model can only do early fusion')
        if input_stats is None:
            input_stats = {}
        if len(input_stats):
            if chan_keys != list(input_stats.keys()):
                # Backwards compat for older pre-fusion input stats method
                assert 'mean' in input_stats or 'std' in input_stats
                input_stats = {
                    chan_keys[0]: input_stats,
                }
            if len(input_stats) != 1:
                print('GOT input_stats = {!r}'.format(input_stats))
                raise ValueError('this model can only do early fusion')
            main_input_stats = ub.peek(input_stats.values())
        else:
            main_input_stats = {}

        if input_stats:
            main_input_stats = ub.peek(input_stats.values())
            self.input_norm = InputNorm(**main_input_stats)
        else:
            self.input_norm = Identity()

        self.backbone = EfficientNet.from_pretrained(MODEL_MAP[network])

        self.n_scales = n_scales

        backbone_level_channels = self.backbone.get_list_features()
        self.neck = BIFPN(in_channels=backbone_level_channels[-n_scales:],
                          out_channels=W_bifpn,
                          stack=D_bifpn,
                          num_outs=5)

        # Note which classes correspond to ignore categories
        self.ignore_class_idxs = []
        for idx, cname in enumerate(self.classes.idx_to_node):
            if cname.lower() == 'ignore':
                self.ignore_class_idxs.append(idx)

        max_level = len(backbone_level_channels)
        pyramid_levels = list(range(max_level - n_scales + 1, max_level + 1))
        self.anchors = Anchors(pyramid_levels=pyramid_levels)

        # TODO: need to parametarize based on anchor spec
        # or not, none of those params are used
        self.bbox_head = RetinaHead(num_classes=num_classes,
                                    in_channels=W_bifpn,
                                    num_anchors=self.anchors.num_per_cell)

        self.regressBoxes = BBoxTransform()  # TODO: move to coder
        # self.clipBoxes = ClipBoxes()  # TODO: move to coder

        if 0:
            for m in self.modules():
                if isinstance(m, nn.Conv2d):
                    n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                    m.weight.data.normal_(0, math.sqrt(2. / n))
                elif isinstance(m, nn.BatchNorm2d):
                    m.weight.data.fill_(1)
                    m.bias.data.zero_()
            self.freeze_bn()

        self.criterion = FocalLoss()

        self.threshold = threshold
        self.iou_threshold = iou_threshold
        self.coder = EfficientDetCoder(
            self.classes, self.threshold, self.iou_threshold)

    def _encode_batch(self, batch):
        """
        Transform bioharn inputs into efficientdet format
        """
        if not isinstance(batch, dict):
            raise Exception('expected bioharn style batch')

        if isinstance(batch['inputs'], dict):
            assert len(batch['inputs']) == 1, ('only early fusion for now')
            inputs = ub.peek(batch['inputs'].values())
        else:
            inputs = batch['inputs']
        assert len(inputs.data) == 1
        imgs = inputs.data[0]

        device = self.backbone._conv_stem.weight.device
        imgs = imgs.to(device)

        if 'label' in batch:
            label = batch['label']
            batch_size = imgs.shape[0]

            if 'tlbr' in label:
                assert len(label['tlbr'].data) == 1
                tlbr_tensor = padded_collate(label['tlbr'].data[0], -1)
                boxes_tensor = kwimage.Boxes(tlbr_tensor, 'tlbr')
            else:
                assert len(label['cxywh'].data) == 1
                cxywh_tensor = padded_collate(label['cxywh'].data[0], -1)
                boxes_tensor = kwimage.Boxes(cxywh_tensor, 'cxywh')

            tlbr_tensor = boxes_tensor.to_tlbr().data
            cidx_tensor = padded_collate(
                [c[:, None] for c in label['class_idxs'].data[0]], -1)
            annotations = torch.cat([tlbr_tensor, cidx_tensor.float()], axis=2)
            annotations = annotations.view(batch_size, -1, 5)
            annotations = annotations.to(device)
        else:
            annotations = None

        return imgs, annotations

    def forward(self, batch, return_result=None, return_loss=None):
        imgs, annotations = self._encode_batch(batch)

        imgs = self.input_norm(imgs)

        # x - a tuple of feature pyramid layers
        pyramid_feats = self.extract_feat(imgs)

        # get class scores / boxes for each pyramid layer
        outs = self.bbox_head(pyramid_feats)
        cls_score, bbox_pred = outs

        if 0:
            print('pyr = ' + ub.repr2([z.shape for z in pyramid_feats]))
            print('cls = ' + ub.repr2([z.shape for z in cls_score]))
            print('box = ' + ub.repr2([z.shape for z in bbox_pred]))

        classifications = torch.cat([out for out in cls_score], dim=1)
        regressions = torch.cat([out for out in bbox_pred], dim=1)

        # TODO: memoize this
        anchors = self.anchors(imgs.shape[-2:], imgs.device)

        # annotations should be B,5,N shaped tensor with cxywh + cidx label
        # cidx of -1 indicates unused
        # annotations = annotations.permute(0, 2, 1)

        if annotations is not None:
            ignore_idxs = self.ignore_class_idxs
            clf_loss, regression_loss = self.criterion(
                classifications, regressions, anchors, annotations,
                ignore_idxs)
            loss_parts = {
                'clf_loss': clf_loss,
                'regression_loss': regression_loss,
            }
        else:
            loss_parts = None

        transformed_anchors = self.regressBoxes(anchors, regressions)
        # transformed_anchors = self.clipBoxes(transformed_anchors, imgs)

        outputs = {
            'transformed_anchors': transformed_anchors,
            'classifications': classifications,
            'loss_parts': loss_parts,
        }
        return outputs

    def freeze_bn(self):
        '''Freeze BatchNorm layers.'''
        for layer in self.modules():
            if isinstance(layer, nn.BatchNorm2d):
                layer.eval()

    def train(self, mode=True):
        r"""Sets the module in training mode.

        Args:
            mode (bool): whether to set training mode (``True``) or evaluation
                         mode (``False``). Default: ``True``.

        Returns:
            Module: self
        """
        # Don't unfreeze BN layers when training
        self.training = mode
        for module in self.children():
            if not isinstance(module, nn.BatchNorm2d):
                module.train(mode)
        return self

    def extract_feat(self, imgs):
        """
            Directly extract features from the backbone+neck
        """
        bb_feats = self.backbone(imgs)
        pyramid_feats = self.neck(bb_feats[-self.n_scales:])
        if 0:
            print(ub.repr2([z.shape for z in bb_feats]))
            print(ub.repr2([z.shape for z in pyramid_feats]))
        return pyramid_feats




def test_time_sampler_issue1():
    """
    Test an issue where our sampler was sampling the same frame multiple times.
    """
    from geowatch.tasks.fusion.datamodules.temporal_sampling.sampler import TimeWindowSampler
    import numpy as np
    nan = np.nan

    kwargs = {
        'time_kernel': np.array([-31536000. ,         0. ,   2629756.8,  31536000. ]),
        'sensors': [None, None, None, None, None, None, None, None, None, None],
        'unixtimes': np.array([nan, nan, nan, nan, nan, nan, nan, nan, nan, nan]),
        'time_window': 4,
        'update_rule': 'distribute',
        'affinity_type': 'uniform',
        'deterministic': False,
        'gamma': 1,
        'name': 'toy_video_2',
        'time_span': None,
        'allow_fewer': True,
    }

    sampler = TimeWindowSampler(**kwargs)
    sampled_idxs, info = sampler.sample(0, return_info=True)

    # There was a bug where we had samples returning the same index multiple
    # times, which should not be allowed. Test for this.
    assert len(sampled_idxs) == len(set(sampled_idxs)), (
        'Sampled indexes are not unique!'
    )



from __future__ import annotations

from typing import Any, Iterable, Mapping, TypeVar
from collections.abc import Callable

import torch
from torch import nn
from torch.nn import functional as F
from torch._C import ScriptMethod

from .computation_node import ModuleNode, FunctionNode, TensorNode, NodeContainer
from .computation_graph import ComputationGraph

from .utils import OrderedSet

# Needed for module wrapper and resetting
_orig_module_forward = torch.nn.Module.__call__


# Functions below are torch.tensor creation operations
# Inside Recorder context, we wrapped these functions
# so that these return RecorderTensor can be recorded during
# forward propagation
orig_name_list = [
    "as_tensor", "from_numpy", "zeros", "zeros_like",
    "ones", "ones_like", "arange", "range", "linspace",
    "logspace", "eye", "empty", "empty_like", "full",
    "full_like", "complex", "heaviside", "bernoulli",
    "multinomial", "normal", "poisson", "rand", "rand_like",
    "randint", "randint_like", "randn", "randn_like",
    "randperm"
]
_orig_op_list = [getattr(torch, name) for name in orig_name_list]


class Recorder:
    '''Context Manager that sets modules forward and torch creation ops
    to record them in computation graph'''
    def __init__(
        self, orig_mod_forward: Callable[..., Any], new_mod_forward: Callable[..., Any],
        model_graph: ComputationGraph
    ) -> None:
        self.orig_module_forward = orig_mod_forward
        self.new_module_forward = new_mod_forward
        self.model_graph = model_graph

    def __enter__(self) -> None:
        setattr(
            torch.nn.Module, "__call__", self.new_module_forward
        )

        for name, op in zip(orig_name_list, _orig_op_list):
            setattr(
                torch, name, creation_ops_wrapper(op, self.model_graph)
            )

    def __exit__(self, exc_type: Any, exc_value: Any, exc_traceback: Any) -> None:
        # reset module __call__ back to original method and torch creation ops
        setattr(
            torch.nn.Module, "__call__", self.orig_module_forward
        )

        for name, op in zip(orig_name_list, _orig_op_list):
            setattr(
                torch, name, op
            )


def creation_ops_wrapper(
    _orig_op: Callable[..., Any], model_graph: ComputationGraph
) -> Callable[..., Any]:
    def _func(*args: Any, **kwargs: Any) -> RecorderTensor:

        input_tensor = _orig_op(*args, **kwargs)
        current_depth = model_graph.context_tracker['current_depth']
        current_context = model_graph.context_tracker['current_context']

        input_recorder_tensor: RecorderTensor = input_tensor.as_subclass(RecorderTensor)
        input_node = TensorNode(
            tensor=input_recorder_tensor,
            depth=current_depth,  # type: ignore[arg-type]
            name='input-tensor' if current_depth == 0 else 'hidden-tensor',
            context=current_context
        )
        current_context.append(input_node)  # type: ignore[attr-defined]
        input_recorder_tensor.tensor_nodes = [input_node]

        return input_recorder_tensor
    return _func


def module_forward_wrapper(model_graph: ComputationGraph) -> Callable[..., Any]:
    '''Wrapper for forward functions of modules'''
    def _module_forward_wrapper(mod: nn.Module, *args: Any, **kwargs: Any) -> Any:
        '''Forward prop of module for RecorderTensor subclass
        Construct Module Node => forward-prop => process output nodes to retain
        module hierarchy correctly
        '''
        # Create module node and connect to its parents tensor node
        input_nodes: NodeContainer[TensorNode] = (
            reduce_data_info([args, kwargs], collect_tensor_node, NodeContainer())
        )
        # get unique input tensors, prevent duplications for auxiliary nodes
        input_recorder: OrderedSet[RecorderTensor] = (
            reduce_data_info([args, kwargs], collect_tensor, OrderedSet())
        )
        # if none of args originated from input
        # hence only torch.Tensor
        if not input_nodes:
            return _orig_module_forward(mod, *args, **kwargs)

        # Create module_node and connect to its parents tensor node
        cur_depth = next(iter(input_nodes)).depth
        input_context = next(iter(input_nodes)).context
        cur_node = ModuleNode(
            mod, cur_depth, input_nodes,  # type: ignore[arg-type]
            name=type(mod).__name__
        )
        cur_node.set_input_shape(
            reduce_data_info([args, kwargs], collect_shape, [])
        )

        # update context with current modules's context
        input_context.append({cur_node: []})
        for node in input_nodes:
            node.add_child(cur_node)

        tensor_to_node: dict[RecorderTensor, TensorNode] = (
            reduce_data_info([args, kwargs], collect_tensor_node_id_dict, {})
        )
        attach_kwargs = {
            'parents': cur_node, 'depth': cur_depth+1,
            'context': input_context[-1][cur_node], 'is_aux': True,
            'name': 'auxiliary-tensor'
        }

        traverse_data_inplace(
            input_recorder, attach_node(attach_kwargs, tensor_to_node)
        )

        model_graph.context_tracker['current_depth'] = cur_depth+1
        model_graph.context_tracker['current_context'] = input_context[-1][cur_node]

        # TODO: check if output contains RecorderTensor
        # this seems not to be necessary so far
        out = _orig_module_forward(mod, *args, **kwargs)

        model_graph.context_tracker['current_depth'] = cur_depth
        model_graph.context_tracker['current_context'] = input_context

        # pop appropriate nodes, see implementation below
        output_recorder: OrderedSet[RecorderTensor] = (
            reduce_data_info(out, collect_tensor, OrderedSet())
        )

        traverse_data_inplace(
            output_recorder,
            process_output_node(cur_node)
        )

        traverse_data_inplace(
            input_recorder, pop_after_forward, recorded_output=output_recorder,
        )

        # remove auxiliary tensor nodes from recorder_tensor
        output_nodes: NodeContainer[TensorNode] = (
            reduce_data_info(out, collect_tensor_node, NodeContainer())
        )

        for output_node in output_nodes:
            cur_node.add_output_nodes(output_node)
            output_node.context = input_context

        cur_node.set_output_shape(reduce_data_info(out, collect_shape, []))
        return out

    return _module_forward_wrapper


class RecorderTensor(torch.Tensor):
    '''Subclass of torch.Tensor used for constructing visual computation graph.

    This class stores list of TensorNode objects to keep record of Nodes during forward
    propagation. The torch_function is also overriden to record needed nodes for visual
    computation graph.

    Attributes:
        All the inherited attributes from torch.Tensor
        tensor_nodes: list[TensorNode]
            List of TensorNode objects to store relevant TensorNodes'''
    @staticmethod
    def __new__(
        cls: Any,
        x: Any,
        tensor_nodes: Any,
        *args: Any,
        **kwargs: Any
    ) -> Any:
        # pylint: disable=unused-argument
        return super().__new__(cls, x, *args, **kwargs)  # type: ignore[call-arg]

    def __init__(
        self, x: Any, tensor_node: TensorNode | list[TensorNode]
    ):
        # pylint: disable=unused-argument
        # super().__init__() # optional

        if isinstance(tensor_node, TensorNode):
            self.tensor_nodes = [tensor_node]
        else:
            self.tensor_nodes = tensor_node

    @classmethod
    def __torch_function__(
        cls: Any, func: Callable[..., Any] | ScriptMethod,
        types: Any,
        args: Any = (),
        kwargs: Any = None,
    ) -> Any:
        '''Calls torch functions for RecorderTensor subclass of torch.Tensor
        Forward prop => Construct Function Node => Construct Output TensorNode
        Args:
            The same arguments as that of  original __torch_function__
            except that the tensor that originated from input (through forward prop)
            are RecorderTensors
        '''
        if kwargs is None:
            kwargs = {}

        args_nodes: NodeContainer[TensorNode] = (
            reduce_data_info([args, kwargs], collect_tensor_node, NodeContainer())
        )

        # This is necessary for torch version < 1.10
        if func in [F.linear, F.embedding]:
            out = nn.parameter.Parameter.__torch_function__(
                func, types, args, kwargs).as_subclass(RecorderTensor)
        else:
            # use original torch_function; otherwise,
            # it leads to infinite recursive call of torch_function
            out = super().__torch_function__(func, types, args, kwargs)

        # if no RecorderTensor is found in input or output
        # dont create any node, give the result only
        if not args_nodes:
            return out
        if not reduce_data_info(out, collect_tensor, OrderedSet()):
            return out

        # Create function_node and connect to its parents tensor node
        cur_depth = next(iter(args_nodes)).depth
        input_context = next(iter(args_nodes)).context
        func_name = (
            func.name if isinstance(func, ScriptMethod) else func.__name__
        )
        cur_node = FunctionNode(
            func, cur_depth, args_nodes, name=func_name  # type: ignore[arg-type]
        )

        for i in args_nodes:
            i.add_child(cur_node)

        input_context.append(cur_node)
        attach_kwargs = {
            'parents': cur_node, 'depth': cur_depth, "context": input_context,
            'is_aux': False, 'parent_hierarchy': {cur_depth: cur_node},
            'name': 'output-tensor' if cur_depth == 0 else 'hidden-tensor'
        }
        traverse_data_inplace(out, attach_node(attach_kwargs))

        # note that when processing inplace operation, input shape is calculated
        # correctly only if inplace operation preserves the input shape
        # which it does for all torch-builtin inplace operations
        # you cant use this before output computation since shape calls
        # to another torch_function (infinite recursion)
        cur_node.set_input_shape(
            reduce_data_info([args, kwargs], collect_shape, [])
        )
        cur_node.set_output_shape(reduce_data_info(out, collect_shape, []))

        return out


L = TypeVar('L', bound=Iterable[Any])


def reduce_data_info(
    recorded_data: Any, action_fn: Callable[..., Any], collected: L, **kwargs: Any
) -> L:
    '''Apply action_fn to RecorderTensor inside recorded_data to collect info of
    input data into collected (Iterable) e.g. shape of RecorderTensor'''
    if isinstance(recorded_data, RecorderTensor):
        action_fn(recorded_data, collected, **kwargs)
    elif isinstance(recorded_data, Mapping):
        for r_d in recorded_data.values():
            reduce_data_info(r_d, action_fn, collected, **kwargs)
    elif (
        isinstance(recorded_data, Iterable) and
        not isinstance(recorded_data, (str, torch.Tensor))
    ):
        for r_d in recorded_data:
            reduce_data_info(r_d, action_fn, collected, **kwargs)
    return collected


def traverse_data_inplace(
    recorded_data: Any, action_fn: Callable[..., Any], **kwargs: Any
) -> None:
    '''Apply action_fn RecorderTensor objects inside recorded_data to change data
    Usuall action_fn is a function that transforms RecorderTensor in memory'''
    if isinstance(recorded_data, RecorderTensor):
        action_fn(recorded_data, **kwargs)
    elif isinstance(recorded_data, Mapping):
        for r_d in recorded_data.values():
            traverse_data_inplace(r_d, action_fn, **kwargs)
    elif (
        isinstance(recorded_data, Iterable) and
        not isinstance(recorded_data, (str, torch.Tensor))
    ):
        for r_d in recorded_data:
            traverse_data_inplace(r_d, action_fn, **kwargs)


def attach_node(
    kwargs: dict[str, Any],
    tensor_to_node: dict[RecorderTensor, TensorNode] | None = None
) -> Callable[..., Any]:
    '''Creates the function to attach TensorNodes, needed for nested calls'''
    def _func(recorded_tensor: RecorderTensor) -> None:
        '''Attaches TensorNode to ModuleNode or FunctionNode
        '''

        if kwargs['is_aux'] and tensor_to_node:
            kwargs['main_node'] = tensor_to_node[recorded_tensor]
        new_kwargs = {
            key_word: value
            for key_word, value in kwargs.items() if key_word != 'tensor_to_node'
        }
        tensor_node = TensorNode(
            tensor=recorded_tensor,
            **new_kwargs
        )
        if isinstance(kwargs["parents"], ModuleNode):
            assert getattr(recorded_tensor, 'tensor_nodes', None) is not None, (
                f'RecorderTensor to be attached to the Node'
                f'{kwargs["parents"]} must have tensor node'
            )
        assert isinstance(kwargs["parents"], (FunctionNode, ModuleNode)), (
            f'Node {kwargs["parents"]} to which to attach must be either'
            f'FunctionNode or ModuleNode'
        )

        if getattr(recorded_tensor, 'tensor_nodes', None) is None:
            recorded_tensor.tensor_nodes = [tensor_node]
        else:
            # ModuleNode: Attaches auxiliary node to tensors
            # Auxiliary nodes should be appended to keep track the node
            # history of tensor
            # FunctionNode: These should overwrite the last tensor node
            # There are 2 different cases:
            # Non-inplace ops -> New tensor and this node is the first
            # Inplace ops -> Result tensor overwrites the intput tensor
            # in memory, so it should overwrite in node history as well
            # for both cases, overwritting the last tensor node is correct
            if isinstance(kwargs["parents"], ModuleNode):
                recorded_tensor.tensor_nodes.append(tensor_node)
            elif isinstance(kwargs["parents"], FunctionNode):
                recorded_tensor.tensor_nodes[-1] = tensor_node
        kwargs["parents"].add_child(tensor_node)
        kwargs['context'].append(tensor_node)
    return _func


def pop_after_forward(
    r_in: RecorderTensor,
    recorded_output: OrderedSet[RecorderTensor],
) -> None:
    '''Removes/pops nodes from RecorderTensors to maintain correct nodes
    Two types of process exist for types of modules:
    Non-inplace ops => pop auxiliary nodes
    In-place ops => pop input nodes since inplace ops overwrites input in memory.
    '''

    in_place_func_message = (
        'Tensor before and after inplace operation must have the same memory address'
    )
    output_id: OrderedSet[int] = OrderedSet(id(x) for x in recorded_output)

    if id(r_in) not in output_id:
        _ = reduce_data_info(
            r_in, collect_tensor_node, NodeContainer(), is_pop=True
        )

    # input of inplace operation
    else:
        assert id(r_in) == r_in.tensor_nodes[-1].tensor_id, (
            in_place_func_message
        )
        assert id(r_in) == r_in.tensor_nodes[-2].tensor_id, (
            in_place_func_message
        )
        # pop tensor node before inplace operation
        r_in.tensor_nodes.pop(-2)


def collect_tensor_node(
    recorded_data: RecorderTensor,
    collected: NodeContainer[TensorNode],
    is_pop: bool = False,
) -> None:
    if getattr(recorded_data, 'tensor_nodes', None):
        if is_pop:
            collected.add(recorded_data.tensor_nodes.pop())
        else:
            collected.add(recorded_data.tensor_nodes[-1])


def collect_tensor_node_id_dict(
    recorded_data: RecorderTensor,
    collected: dict[RecorderTensor, TensorNode],
) -> None:
    if getattr(recorded_data, 'tensor_nodes', None):
        collected[recorded_data] = recorded_data.tensor_nodes[-1].main_node


def collect_tensor(
    recorded_data: RecorderTensor, collected: OrderedSet[RecorderTensor]
) -> None:
    collected.add(recorded_data)


def collect_shape(
    recorded_data: RecorderTensor, collected: list[tuple[int, ...]]
) -> None:
    collected.append(tuple(recorded_data.shape))


def process_output_node(
    cur_node: ModuleNode
) -> Callable[..., Any]:
    '''Returns function to update output node after forward
    pass of nn.Modules'''
    def _func(recorded_data: RecorderTensor) -> None:
        output_node = recorded_data.tensor_nodes[-1]
        cur_depth = cur_node.depth
        # if output node is reused inside module or empty module is used
        # introduce node for empty pass function
        if not output_node.is_leaf() or output_node.is_aux:
            insert_empty_pass_node(recorded_data, output_node)

        recorded_data.tensor_nodes[-1].depth = cur_depth
        name = 'output-tensor' if cur_depth == 0 else 'hidden-tensor'
        recorded_data.tensor_nodes[-1].name = name
        recorded_data.tensor_nodes[-1].parent_hierarchy[cur_depth] = cur_node
    return _func


def insert_empty_pass_node(
    recorded_tensor: RecorderTensor, out_node: TensorNode
) -> None:
    '''First, inserts empty-pass node as a child of tensor nodes. Then, inserts
    TensorNode as a child of this empty-pass node'''
    out_pass = FunctionNode(
        lambda x: x, out_node.depth, out_node,
        name='empty-pass'
    )
    out_node.add_child(out_pass)
    out_node.context.append(out_pass)

    passed_out_node = TensorNode(
        recorded_tensor, out_node.depth, out_pass,
        context=out_node.context, is_aux=False,
        parent_hierarchy={
            recorded_tensor.tensor_nodes[-1].depth: out_pass
        }
    )

    out_node.context.append(passed_out_node)
    out_pass.add_child(passed_out_node)

    # Update the current node of RecorderTensor
    # Here append instead of overwrite the last node because
    # this is a dummy FunctionNode that has no actual place in
    # computation graph
    recorded_tensor.tensor_nodes.append(passed_out_node)
